"""Auxiliary functions for writing tables into C program"""

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import types
import sys
import re
import os
import warnings
import itertools
import ast
from io import IOBase
from operator import __or__, __xor__, __and__
from numbers import Integral
from collections.abc import Iterable
from functools import partial
import numpy

import regex

if sys.version_info[0] >= 3:
    file = IOBase 
    long = int    



#######################################################################
#######################################################################
# Functions provided for the user of the code generator
#######################################################################
#######################################################################

#######################################################################
# Auxiliary functions
#######################################################################

def isinteger(x):
    """Return True if x is an integer"""
    return  isinstance(x, (Integral, numpy.integer))


#######################################################################
# Formatting a python object for use in a C table
#######################################################################


int_format_dict = {8 : (2,""), 16:(4,""), 32:(8,"L"), 64:(16,"LL")}
#  32:(8,"L") means: format "uint32" contains 8 hex digits and suffix "L"

m_int_format = re.compile(r"(u?)int([0-9]+)")

def format_item(n, format = None):
    """Format a python object ``n`` to a C string for use in a C table

    The default format is to apply function str() to the object.

    The current version supports integer formats only.
    Parameter ``format`` must be one of the following strings:

     ``'int8', 'int16', 'int32', 'int64', 'uint8', 'uint16', 'uint32', 'uint64'``
    """
    if not format:
        s = str(n)
        if isinteger(n) and s[-1]=='L':
            s = s[:-1]
        return s
    m =  m_int_format.match(format)
    if m:
        blen =  int(m.group(2))
        n = int(n) 
        try:
            len, suffix = int_format_dict[blen]
        except:
            raise ValueError("Bad integer format " + str(format))
        if m.group(1) == 'u':
            n &=  (1 << blen) - 1
            if suffix: suffix = "U" + suffix
            return "0x{0:0{1}x}{2}".format(n , len, suffix)
        else:
            return  "{1}0x{0:0{2}x}{3}".format(abs(n),"-"[n>=0:], len, suffix)
    raise ValueError("Bad format " + str(format))



#######################################################################
# Support for user-defined directives
#######################################################################

def prepend_blanks(string, length):
    """Prepend ``length`` blank characters to each line of a ``string``

    Here parameter ``string`` is a string that may consist of
    several lines.
    """
    def prepend_blank_line(line, length):
        return " " * length + line if len(line.strip()) else line
    l = [prepend_blank_line(line, length) for line in string.split("\n")]
    return "\n".join(l)




class UserDirective(object):
    """Convert a function for use in a code generator directive

    Each code generator directive corresponds to a function. Evaluating
    that function with the arguments given to the directive results in
    a string that makes up the code generated by the directive.

    The list of arguments of a code generator directive is necessarily 
    a string. Usually, this string is converted to a python expression
    which is evaluated using the dictionary 'tables' given to the code
    generator. Here dictionary ``tables`` contains the local variables 
    used by the evaluation process. 

    The constructor of this class returns an object which is a callable 
    function. Calling that function has the same effect as calling the
    function ``f``, which is the first argument passed to the 
    constructor.

    Additional information for dealing with the arguments of the
    function may be given the constructor. This information is used
    by the code generator if an instance of this class is registered
    at the code generator as a directive. For registering a directive
    in that way you may simply enter an instance of this class as a 
    value in the dictionary *directives* passed to the constructor of 
    class ``TableGenerator``.

    Using this feature you may automatically convert arguments of the 
    function e.g. to strings, integers or python objects. You may also 
    prepend a local variable from dictionary ``tables`` to the list 
    of user-specified arguments of the directive. The latter feature 
    gives a directive the flavour of a member function of a class, 
    similar to prepending ``self`` to the list of user-specified 
    arguments.

    :param f:
   
        A callable function which is executed when the code generator
        parses a user-defined directive. The arguments of such a 
        directive are passed to function ``f`` as arguments.

    :param eval_mode:

        This  is a string specifying the syntax of the arguments.
        The i-th character of the string ``eval_mode`` controls the 
        evaluation of the i-th argument of function ``f`` as follows:

          * ``'p'``   Evaluate the argument to a python object (default).
          * ``'i'``   Evaluate the argument to an integer.
          * ``'s'``   Evaluate the argument to a string, ignoring dictionary

          It ``eval_mode`` is ``'.'`` then a single parameter containing 
          just  the given string is passed to function ``f`` as a single 
          argument without any parsing with python.
 
    :param prefix:

        Describes to an optional local variable of the code generator
        prepended to the list of arguments of the directive. It may be
        one of the following:

         *  ``False``:    Nothing is prepended
         *  ``True``:     The dictionary of all local variables is prepended
         * any string:    The local variable with that name is prepended 

    """
    def __init__(self, f, eval_mode = "", prefix = False):
        """Create an instance of class UserDirective for function f 

        Calling that instance with a single argument s of type str has  
        essentially the same effect as calling f with the arguments 
        obtained by  evaluating the string s as a python expression. 
        
        When  registering that instance as a code generator directive,
        the other arguments have the following effect: 

        Remark:
          A dictionary d of local variables provided by the code generator 
          always contains an entry 'NAMES' with  d["NAMES"] = d.  
        """
        if callable(f):
            self.function = f
        else:
            raise TypeError("Function registered for code generator is not callable")          
        self.eval_mode = eval_mode
        if not prefix:
            self.prefixes = ()
        elif isinstance(prefix, str):
            self.prefixes = (prefix,) 
        elif prefix == True:
            self.prefixes = ("NAMES",) 
        else:
             raise TypeError("Prefix must be True, False or a string")
    def set_code_generator(self, code_generator):
        self.tg = code_generator
    def __call__(self, args, *_):
        indent = max(0, self.tg.current_line.find("/"))
        f_args = [self.tg.names[name] for name in self.prefixes]
        f_args += eval_codegen_args(self.tg.names, args, self.eval_mode)
        res = str(self.function(*f_args))
        return prepend_blanks(res, indent),  ""

#######################################################################
# Support for user-defined formatting functions
#######################################################################



class UserFormat(UserDirective):
    r"""Convert a function for string formatting in the code generator

    The code generator evaluates expressions in curly braces ``%{...}``
    in the source code with the python ``.format()`` function for
    stings.

    It is desirable that ``%{int:2*3}`` evaluates to a string 
    representing the integer constant ``int(2*3)`` for some built-in 
    and user-defined functions such as function ``int``.

    Therefore we need a python object with name ``int`` that behaves 
    like the function ``int()`` when being called. That object should 
    also have a ``__format__`` method that returns 
    ``str(int(<expression>))``
    if a string representing a valid python expression is passed to
    the ``__format__`` method.

    ``UserFormat(int)`` creates such an object. Thus:: 

       tables = {"int": UserFormat(int)}
       "{int:2*3}".format(**tables))

    evaluates to ``str(int(2*3))``. Of course, function ``int`` may be
    replaced by any user-defined function.

    Things get more interesting if we prepare a dictionary for the code 
    generator::
    
      from mmgroup.generate_c import TableGenerator, UserFormat
      # Prepare tables for a code generator
      tables = {"int": UserFormat(int),  "a": 3, "b":5}
      # Create a code generator 'codegen'
      codegen = TableGenerator(tables)
      # Make a sample source line of C code
      sample_source = "  x = %{int: a * b + int('1')};\n"
      # generate file "test.c" containing the evaluated sample line
      codegen.generate(sample_source, "test.c")

    Here the string ``{int: a * b + int('1')}`` in the sample source 
    line is formatted with the standard python ``.format`` method.
    The ``.format`` method obtains the entries of the dictionary 
    ``tables`` as keyword arguments.

    So ``%{int: a * b + int('1')}`` evaluates to ``str(3*5+1)``,
    and the C file ``"test.c"`` will contain the line::

      x = 16;


    So ``UserFormat(f)`` creates an object the behaves like ``f`` when 
    it is called; and it returns ``str(f(arg1,arg2,...))`` when it is 
    invoked  in ``"{f:arg1,arg2,...}".format(f = f)``.

    Here the argument list ``arg1, arg2,...`` is processed in according 
    to the same rules as the argument list of a user-defined directive
    in class ``UserDirective``.

    :param f:
   
        A callable function which is executed when the code generator
        parses a user-defined formatting string in curly braces 
        preceded by a '%' character, such as ``%{...}``. 
        The arguments given in the formatting string
        are passed to function ``f`` as arguments.

    :param eval_mode:

        This  is a string specifying the syntax of the arguments in the
        same way as in class ``UserFunction``.
 
    :param prefix:

        Describes to an optional local variable of the code generator
        prepended to the list of arguments of the function in the
        same way as in class ``UserFunction``.

    :param fmt:

        By default, function str() is applied to the result of 
        function ``f`` in order to convert that result to a string.
        This behaviour can be modified by setting parameter ``fmt``.

        if ``fmt`` is a callable function  then ``str(f(*args))`` is 
        replaced by  ``fmt(f(*args))``. 

    :param arg_fmt:

        If this parameter is set then the corresponding ``.format`` 
        method returns ``arg_fmt(*args)`` instead of ``str(f(*args))``.
        This overwrites the parameter ``fmt``.
    """
    def __init__(self, f, eval_mode = "", prefix = False, fmt = None, arg_fmt = None):
        """Create an instance of class UserDirective for function f 

        Calling that instance has the same effect as calling f, except
        that the given arguments for f may be prepended by the another
        argument as specified by  parameter 'prefix' in the 
        constructor. 

        Here parameter 'prefix' specifies an optional prefix to
        be prepended to the list of arguments passed to function f
        as in Class UserDirective.  

        The instance of class UserDirective created here als has a
        __format__method that returns str(f(<expression>))  if a string 
        representing a valid python expression is passed to the 
        __format__ method. Here the expression string is evaluated
        using parameter 'eval_mode' and prepended by an optional
        argument as specified by parameter 'prefix' exactly in the 
        same way as in class UserFormat.

        By default, function str() is applied to the result of 
        function f in order to convert that result to a string.
        This behaviour can be modified by setting one of the
        parameters 'fmt' or 'arg_fmt'.

        if 'fmt' is set then str(f(*args)) is replaced by
        fmt(f(*args)). If 'arg_fmt'  is set then str(f(*args)) 
        is replaced by arg_fmt(*args).
        """
        super(UserFormat, self).__init__(f, eval_mode, prefix) 
        if not arg_fmt is None:
            self.format_function =  arg_fmt
        else:
            if fmt is None: fmt = str
            def fmt_function(*args, **kwds):
               return fmt(f(*args, **kwds))
            self.format_function =  fmt_function
    def __call__(self, *args):
        f_args = [self.tg.names[name] for name in self.prefixes]
        f_args += args
        return self.function(*f_args)
    def __format__(self, format_):
        args = [self.tg.names[name] for name in self.prefixes]
        #print("FORMAT", ":",  format_, ";", self.eval_mode, self.prefixes)
        #print("NAMES", list(self.tg.names.keys()))
        args += eval_codegen_args(self.tg.names, format_, self.eval_mode)
        return self.format_function(*args)



#######################################################################
# Built-in functions 
#######################################################################

# Here is a dictionary of built-in functions that are considered safe
safe_locals = {
    'abs': abs, 'max': max, 'min':min,  'str': str,
    'len': len, 'pow': pow, "int": int, 'range':range,
    'zip': zip,
}



#######################################################################
# Built-in formatting functions
#######################################################################


# Auxiliary functions used in built-in format functions

def strip_L(f):
    """Return a function converting an integer to a string 

    The resulting function uses function f (which should be hex or 
    str) for conversion. Some problems with python 2 are resolved.
    """
    def wrapper(*args, **kwds):
        result = f(*args, **kwds)
        while result[-1:] == 'L': result = result[:-1]
        return result
    return wrapper    


def join_(str_, iterable):
    """functional version of method join of class str

    """
    return str_.join(map(str,iterable))


# Here is the list of built-in format functions
built_in_formats = {
       "int" : UserFormat(int, fmt = strip_L(str)),
       "hex" : UserFormat(lambda x: x, fmt = strip_L(hex)),
       "len" : UserFormat(len, fmt = strip_L(str)),
       "str" : UserFormat(str),
       "join": UserFormat(join_),
}

# Add safe_locals to built_in_formats
for name, f in safe_locals.items():
    if not name in built_in_formats:
        built_in_formats[name] = f

#######################################################################
#######################################################################
# Functions for internal use by the code generator
#######################################################################
#######################################################################






#######################################################################
# Evaluating a string to a python expression using python module ast
#######################################################################


# For background on module ast (abstact syntax tree) see e.g.:
# https://greentreesnakes.readthedocs.io/en/latest/tofrom.html







# Set of nodes in module .ast considered as safe for function eval()
# See  https://greentreesnakes.readthedocs.io/en/latest/nodes.html
good_ast_nodes = set([
   ast.Expression,
   ast.Num, ast.Str, ast.List, ast.Ellipsis, ast.Expr,
   ast.Load,
   ast.Constant,
   ast.Tuple,

   ast.UnaryOp,
   ast.UAdd, ast.USub, ast.Not, ast.Invert,

   ast.BinOp,
   ast.Add, ast.Sub, ast.Mult, ast.Div, ast.FloorDiv, ast.Mod, 
   ast.Pow, ast.LShift, ast.RShift, ast.BitOr, ast.BitXor, 
   ast.BitAnd,

   ast.BoolOp,
   ast.And, ast.Or,

   ast.Compare,
   ast.Eq, ast.NotEq, ast.Lt, ast.LtE, ast.Gt, ast.GtE, ast.Is, 
   ast.IsNot, ast.In, ast.NotIn,

   ast.Call,
   ast.Subscript, ast.Index, ast.Slice, ast.ExtSlice, 
   ast.ListComp, ast.comprehension
])



try:
    good_ast_nodes.add(ast.NameConstant) # New in version 3.4
except:
    pass 


class EvalNodeVisitor(ast.NodeVisitor):
    """Node visitor for Abstract Syntax Trees creeated by module  .ast 

    This vistor checks for all nodes of an AST whether they are secure 
    for processing with function eval(). Method visit() raises TypeError
    if an insecure node id found in the AST.

    Caution: 
    Leading underscores in python identifiers or attributes are illegal! 

    Note that leading underscores in an identifier or an attribute are
    considered dangerous. E.g.  parsing '().__class__.__bases__[0]'
    yields the class <object>, which a user should not manipulate. For  
    a discussion about the security of eval() see:
    https://nedbatchelder.com/blog/201206/eval_really_is_dangerous.html

    Creating tuples in backets or dictionaries in curly brackets is also
    illegal because the syntax gets messy when used in the code generator.
    """
    def visit(self, Node):
        cls = Node.__class__
        #print("visiting", Node, cls)
        if cls in good_ast_nodes:
            pass
        elif cls == ast.Name:
            if Node.id[:1] == "_":
                raise NameError("Illegal leading '_' in name in python expression") 
        elif cls == ast.Attribute:
            if Node.attr[:1] == "_":
                raise NameError("Illegal leading '_' in name in python expression") 
        else:
            print("\nIllegal ast node", type(Node))
            raise TypeError("Illegal construct in python expression") 
        self.generic_visit(Node)


def safe_eval(s,  locals_ = {}):
    """An attempt to create a secure version of function eval()

    See function EvalNodeVisitor() for allowd python contructs.
    """
    a_tree = ast.parse(s.strip(), mode="eval")
    vars_ = {}
    vars_.update(safe_locals)
    vars_.update(locals_)
    EvalNodeVisitor().visit(a_tree)
    code = compile(a_tree, filename="Python expression", mode="eval")
    result = eval(code, {'__builtins__': {}},  vars_ )
    return result


 
   

def ast_untuple(expression):
    """Evaluate string 'expression' to a list of AST nodes.

    If 'expression'is a tuple, a list of AST nodes is returned,
    such that each entry of the list represents an entry of the
    tuple. Otherwise a list with one entry representing the
    'expression' is returned.  
    """
    tree = ast.parse(expression.strip(), mode = "eval")
    if isinstance(tree.body, ast.Tuple):
        arglist = []
        for node in tree.body.elts:
            node = ast.fix_missing_locations(node)
            expr = ast.Expression(node)
            expr = ast.fix_missing_locations(expr)
            arglist.append(expr)   
    else:
        arglist = [tree]
    for arg in arglist:
        EvalNodeVisitor().visit(arg) 
    return arglist
    





#######################################################################
# Evaluating a string to a list of arguments
#######################################################################



def ast_eval_str(node, locals_ = None):
    """Evaluate an AST expression to a string"""
    if isinstance(node.body, ast.Name):
        return node.body.id
    EvalNodeVisitor().visit(node)
    code = compile(node, filename="Python parameter", mode="eval")
    result = str(eval(code, {'__builtins__': {}},  {} ))
    return result
   
    
def ast_eval_pyobj(node, locals_):
    """Evaluate an AST expression to a python object"""
    EvalNodeVisitor().visit(node)
    code = compile(node, filename="Python parameter", mode="eval")
    return eval(code, {'__builtins__': {}},  locals_ )

def ast_eval_int(node,  locals_):
    """Evaluate an AST expression to an int"""
    return int(ast_eval_pyobj(node, locals_))


ast_eval_mode = {
   'p': ast_eval_pyobj,   # evaluate argument to a python object
   'i': ast_eval_int,     # evaluate argument to an int
   's': ast_eval_str,     # evaluate argument to a string
}





   

def eval_codegen_args(tables, expression, eval_mode = ""):
    """Evaluate expression for a user-defined function

    'expression' must be a valid python expression, which will be
    converted to a list of arguments for a user-defined function.

    If 'expression' is not a tuple, a list containing just that
    (evaluated) expression is returned. Otherwise each entry of 
    the tuple is evaluated and the list of those evaluated
    expressions is returned

    'tables' is a list of local variables used for evaluating the
    expression. 

    The i-th character of the string 'eval_mode' controls the
    evaluation of the i-th argument as follows:

      'p'   Evaluate the argument to a python object (default).
      'i'   Evaluate the argument to an integer.
      's'   Evaluate the argument to a string, ignoring dictionary

    It 'eval_mode' is "." then a list containing just the string
    "expression" is returned.
    """
    assert isinstance(expression, str), expression
    #print("eval_codegen_args", tables, expression)
    if eval_mode[:1] == ".":
        return [expression]
    ast_par_list = ast_untuple(expression)
    d = len(ast_par_list)  - len(eval_mode) 
    if d > 0:
        eval_mode += "p" * d
    args = [ ast_eval_mode[fmt](ast_param, tables)
           for (ast_param, fmt) in zip(ast_par_list,  eval_mode)]
    return args
   




#######################################################################
# Updating a dictionary with local variables in FOR directive
#######################################################################

class NotInDict:
    """Models a value for a non-existent key in a dictionary"""
    pass


class SaveDictFor(object):
    """Models shadowing local variables in a FOR directive

    The directive

        // %%FOR <local> in <iterable>

    has the effect that local variables in the list <local> are 
    iteratively set to the values obtained by iterating through
    <iterable>. This shadows the original values of the variables
    with the same names, if present.
    
    If <local> contains only one variable, that variable is set to
    the values yielded by the <iterable>. If <local> is a comma-
    separated list several variables, then <iterable> must yield 
    tuples of exactly the same length as the list <local>, and 
    all local variables are updated accordingly for each tuple
    yielded by the <iterable>. 

    More sophisticated forms of list comprehensions in a FOR
    directive are not supported.

    This constructor of this class takes the dictionary 'dict'
    of current  variables and the argument 'local', which may be a 
    single variable name or an tuple of variable names. It saves
    all items of the dictionary with a key equal to any
    of these names.
   
    Method 'set_local' updates the names all local variables in
    the dictionary using values from an 'entry' onvtained from 
    the <interable>.

    Method 'restore' restores the dictionary of current variables
    to state before calling the constructor.
    """
    def __init__(self, dict, local):
        """Create instance for updating a dictionary of variables

        'dict' is dictionary of variables to be updated.
        'local' is a name of a local variable to be updated;
        it may also be a tuple of such names.
        """
        self.local = local
        self.dict = dict
        self.old_dict = {}
        if not isinstance(local, tuple):
            local = (local,)
        for name in local:
            try: 
                self.old_dict[name] = dict[name]
            except KeyError:
                self.old_dict[name] = NotInDict

    def set_local(self, entry):
        """Set keys in dictionary to values given by 'entry'

        Here 'entry' is an item yielded from the <iterative> in the
        <expression> of the FOR directive. The dictionary of 
        current variables given in the constructor is updated
        apropriately.
        """ 
        if not isinstance(self.local, tuple):
            self.dict[self.local] = entry
        else:
            if len(entry) != len(self.local):
                raise ValueError("Too many or few values to unpack in FOR directive")
            for local, value in zip(self.local, entry):
                self.dict[local] = value
         
    def restore(self):
        """Restore the dictionary passed in the constructor"""
        for name, value in self.old_dict.items():
            if value is NotInDict:
                try:
                    del self.dict[name]
                except:
                    pass
            else:
                self.dict[name] = value
       

         


#######################################################################
# Support for built-in directives of the code generator
#######################################################################


 
def eval_codegen_table(tables, expression):
    """Evaluate 'expression' for table generation

    Here expression is evaluated as a pair (table, format_) where
    'table' is an iterable python object and 'format_' is an optional
    string (default is the empty string).

    'tables' is a list of local variables used for evaluating the
    expression. 

    The function returns a triple (name, table, format_). If parameter
    'table' is a python identifier, then 'name' is set to the name of
    that identifer, otherwise 'name' is set to the empty string.
    """ 
    ast_par_list = ast_untuple(expression)
    node = ast_par_list[0]
    name =  node.body.id if isinstance(node.body, ast.Name) else ""
    table = ast_eval_pyobj(node, tables)
    format_ = ast_eval_str(ast_par_list[1], {}) if len(ast_par_list) > 1 else ""
    return name, table, format_ 
    
    

codegen_assign_match = re.compile(r"\s*([A-Za-z]\w*)\s*=\s*(.+)$")
 
def eval_codegen_assign(tables, expression):
    """Evaluate 'expression' in an assignment function

    Here 'expression' is evaluated must be a string of the form 

        <name> = <value>

    and <value> is evaluated to a python object <py_value>. 

    'tables' is a list of local variables used for evaluating the
    expression. 

    The function returns the pair (<name>, <py_value>).
    """ 
    try:
        name, value = codegen_assign_match.match(expression).groups()
    except:
        raise TypeError("Syntax error in assignment in code generator")   
    return name, safe_eval(value, tables)
    
 


def eval_codegen_m(tables, expression, match_pattern, name):
    """Evaluate 'expression' in a 'FOR' or "WITH" directive.

    Here 'expression'  must "match_pattern" which is of shape 

        '<name> in <value>'    or '<name> = <value>' 

    and <value> is evaluated to a python object <py_value> which 
    should be iterable. 

    'tables' is a list of local variables used for evaluating the
    expression. 

    The function returns the pair (<name>, <py_value>).
    """ 
    try:
        names, value = match_pattern.match(expression).groups()
        names = tuple(s.strip() for s in names.split(","))
        for s in names: 
            assert s[:1].isalpha()
        assert len(names) > 0
        if len(names) == 1: 
            names = names[0]
    except:
        err = "Syntax error in %d directive in code generator"
        raise TypeError(err % name) 
    return names, safe_eval(value, tables)

codegen_assign_for = re.compile(r"\s*([,\w\s]+)\s+in\s+(.+)$")

def eval_codegen_for(tables, expression):
    """Evaluate 'expression' in a 'FOR' directive.

    Here 'expression'  must be a string of the form 

        <name> in <value>

    and <value> is evaluated to a python object <py_value> which 
    should be iterable. 

    'tables' is a list of local variables used for evaluating the
    expression. 

    The function returns the pair (<name>, <py_value>).
    """ 
    return eval_codegen_m(tables, expression, codegen_assign_for, "FOR")

codegen_assign_with = re.compile(r"\s*([,\w\s]+)=(.+)$")

def eval_codegen_with(tables, expression):
    """Evaluate 'expression' in a 'WITH' directive.

    Here 'expression'  must be a string of the form 

        <name> = <value>

    and <value> is evaluated to a python object  <py_value>. If 
    <name> is a tuple of names then <value> should evaluate to
    a tuple <py_value> of the same length. 

    The function returns the pair (<name>, <py_value>). Operation
    is similar to function eval_codegen_for().
    """ 
    return eval_codegen_m(tables, expression, 
        codegen_assign_with, "WITH")




codegen_assign_if = re.compile(r"\s*(IF\s+)?(.+)$")


def eval_codegen_else(tables, expression):
    """Evaluate 'expression' in an 'ELSE' directive.

    Here 'expression'  must be a string of the form 

        IF <value>

    where the 'IF' keyword is optional. <value> is evaluated to a 
    Boolean value <bool_value>. 

    'tables' is a list of local variables used for evaluating the
    expression. 

    The function returns the pair (<bool_value>, <if_found>), where
    <if_found> is True iff the expression contains the IF kewyord.
    """ 
    try:
        if_found, condition = codegen_assign_if.match(expression).groups()
    except:
        if len(expression.strip()) == 0:
            return True, False  # empty ELSE expression
        raise TypeError("Syntax error in IF directive in code generator")
    condition = condition.strip()
    condition = safe_eval(condition, tables) if condition else 1
    #  print("ELSE", condition, ";", if_found)
    return  bool(condition),  bool(if_found)

      
m_direct_arg = re.compile(r"\s*([a-zA-Z0-9_]+)(\s|,|$)")

def direct_single_arg(expression, kwd = ""):
    """Evaluate 'expression' to an alphanumeric string.

    The function returns that string.

    'kwd' is a string indicating the keyword for which this function
    has been called. This is used for error reporting.
    """
    try:
        return m_direct_arg.match(expression).groups()[0]
    except:
        raise TypeError("Syntax error in codegen directive %s" % kwd)
    

def eval_codegen_join(tables, expression):
    """Evaluate 'expression' in a 'JOIN' directive.

    Here 'expression'  must be a string of the form 

        <infix> , <suffix>

    Both, <infix> and  <suffix> are evaluated to a string. If any 
    of the evluated strings does no contain a '\n' character, 
    a '\n' is appended to the string.

    'tables' is a list of local variables used for evaluating the
    expression. 

    The function returns a pair (infix, suffix) of strings.
    """ 
    args = eval_codegen_args(tables, expression, "ss")
    while len(args) < 2: 
        args.append("")
    for i in range(2):
        if not "\n" in args[i]:
            args[i] =  args[i] + "\n"
        if (args[i][:1]).isidentifier():
            args[i] = " " + args[i]
        if (args[i][-1:]).isidentifier():
            args[i] = args[i] + " "
    return tuple(args[:2])

#######################################################################
# Evaluating a balanced list of arguments, deprecated!!!
#######################################################################
     


bracket_match = { ")":"(", "]":"[", "}":"{" }


def balanced_arglist(s):
    """Parse a comma-separated string s into a list of strings
    
    Deprecated!

    Semicolons and unbalanced brackets, square bracktes or curly 
    brackets are illegal. The strings "#" and "//" (outside 
    brackets) mark the beginning  of a comment. 

    The function separates the arguments by a comma and returns
    the list of arguments as strings with leading and trailing
    whitespace stripped off.
    """
    assert isinstance(s, str), s
    brackets = []
    args = []
    start, end_ = 0, len(s)
    for i, c in enumerate(s):
        if c in "([{": 
            brackets.append(c)
        elif c in ")]}":
            try:
                assert  brackets.pop() == bracket_match[c]
            except:
                print("\nOpening/closing bracket mismatch in string:")
                print(s)
                raise ValueError("Opening/closing bracket mismatch")
        elif c == "," and not len(brackets):
             args.append(s[start:i].strip())
             start = i+1
        elif c in ";/#" and not len(brackets):
            if c == "#" or s[i:i+2] == "//":
                 end_ = i
                 break
            elif c == ";":
                print("\nIllegal character %s in string:" % c)
                print(s)
                raise ValueError("String format error")
    last = s[start:end_].strip()
    if len(last):
        args.append(last)
    return args










#######################################################################
# Support for generating tables
#######################################################################







def unformat_item(s, format = None):
    """Convert a C literal string to a python object, 

    Deprecated!!!

    This tries to reverse the effect of function format_item.
    """
    if not format:
        return ast.literal_eval(s)
    m =  m_int_format.match(format)
    if m:
        while len(s) and s[-1] in "UL":
            s = s[:-1] 
        return ast.literal_eval(s)
    raise ValueError("Cannot interpret C literal " + str(s))

def sizeof_format(format = None):
    """Return size of entry with of a given format in bytes

    Parameter format is as in function format_item().
    The function returns 0 if the size cannot be determined.
    """
    try:
        m =  m_int_format.match(format)
        return int(m.group(2))/8
    except:
        return 0


def make_table(table, format=None):
    """Convert a python table to a C sequence separated by commas

    Parameter ``table`` may be anything iterable. Function 

      ``format_item(x, format)``

    is applied to each item ``x`` in the table.

    The output may contain several lines.
    """
    l = 77 // (len(format_item(0,format)) + 1)
    while (l & (l-1)): l = l-1
    s = ""
    for i,x in enumerate(table):
        if i % l == 0: s += ""
        s += format_item(x, format) 
        if i < len(table) - 1: s += ","
        if i % l == l-1: s += "\n"
    if s[-1] != "\n": s += "\n"
    return s    


class named_table(object):
    """Helper class for class TableGenerator for modeling a table
     
    self.str() returns the name of the table
    self[i] returns the i-th entry of the table
    """
    def __init__(self, name, table):
        self.name = name
        self.table = table
    def __repr__(self):
        return self.name
    def __getitem__(self, i):
        return self.table.__getitem__(i)
    def __len__(self):
        return len(self.table)
    str = __repr__


#######################################################################
# Suppport for generating doc strings
#######################################################################


def n_leading_spaces(s):
    """Return number of leading spaces of s"""
    return sum(1 for _ in itertools.takewhile(lambda x : x.isspace(), s))
    #6789012345678901234567890123456789012345678901234567890123456789012345


def py_doc_to_comment(s):
    """Convert python __doc__ string to a C comment

    """
    ls = [s1.rstrip() for s1 in str(s).split("\n")]
    while len(ls) and len(ls[-1]) == 0:
        del ls[-1]
    if len(ls) == 0: return ""
    try:
        trim_len = min(n_leading_spaces(s1) for s1 in ls[1:] if len(s1))
    except ValueError:
        trim_len = 0
    ls = ls[:1] + [s2[trim_len:] for s2 in ls[1:]]
    return   "\n".join("// " + s for s in ls) + "\n"



#######################################################################
# Support for formating a C code line
#######################################################################




# The following expression matches  a string  with balanced curly 
# braces and preceded by a '%' character.
mm_line = regex.compile(r"\%(\{(?>[^{}]|(?1))*\})")


def _do_format(line, format_args, format_kwds, terminal, matchobj):
    try:
        return matchobj.group()[1:].format(*format_args, **format_kwds)
    except:
        if not terminal:
            return matchobj.group()
        print("\nError in formatting line:")
        print(line)
        print("Cannot expand expression", matchobj.group() + "\n")
        raise


def format_line(line, dictionary=None, args = (), terminal = True):
    """Format a C code ``line``

    The function scans  a ``line`` for occurences of a string 
    with balanced curly braces and preceded by a '%' character.
    When such a string ``s``  is found, then ``s`` is substituted
    by ``s[1:].format(*args, **dictionary)``.
    
    We consider this to be a reasonable compromise between the 
    standard C syntax an the very powerful pythonic ``.format``
    method for strings.
    """
    if line.find(r"%{") < 0:
        return line
    f_subst = partial(_do_format, line, args, dictionary, terminal)
    try:
        return regex.sub(mm_line, f_subst, line)
    except:
        print("\nError in formatting line:")
        print(s + "\n")
        raise



#######################################################################
# Indentation of a block
#######################################################################



def indent_subsequent_lines(s):
    if "\n" in s[1:-1]:
        lines = s.split("\n")
        index = 0
        while lines[0][index].isspace(): index += 1
        indent = lines[0][:index]
        for  i, s in enumerate(lines[1:]):
            if len(s): lines[i+1] = indent + s
        return "\n".join(lines)
    else:
        return s


#######################################################################
# Support for generating header files
#######################################################################

def line_as_comment(line):
    """Change a C code line to a comment by adding the prefix "// "

    If the last character of the line is a backslash, certain
    characters are appended. This silences some compiler warnings.
    """
    line = "// " + line
    if line[-2:] == "\\\n": 
        line = line[:-2] +  "\\ //\n"
    return line








"""Test C function dealing with Leech lattice vectors mod 3 of type 3

In this script we test functions dealing with vectors of type 3 in the 
Leech lattice modulo 3. These functions are implemented in file 
gen_leech3.c and available in the extension mmgroup.generators.

We use the terminology defined in
the document *The C interface of the mmgroup project*, 
section *Description of the mmgroup.generators extension*.
"""

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals



from random import randint, choices #, shuffle, sample
from functools import reduce
from operator import __or__
from numbers import Integral
from collections import defaultdict
from multiprocessing import Pool, TimeoutError

import numpy as np
import scipy
import scipy.special
from scipy.stats import chisquare
import pytest

from mmgroup import MM0
from mmgroup.mat24 import MAT24_ORDER, ploop_theta
from mmgroup.mat24 import bw24 as mat24_bw24
from mmgroup.generators import gen_leech3to2_type3
from mmgroup.generators import gen_leech3_op_vector_word
from mmgroup.generators import gen_leech3_op_vector_atom
from mmgroup.generators import gen_leech2_op_word
from mmgroup.generators import gen_leech2_op_atom
from mmgroup.generators import gen_leech2_type_selftest

from mmgroup.tests.test_gen_xi.test_gen_type4 import rand_xsp2co1_elem
from mmgroup.tests.test_gen_xi.test_gen_type4 import create_test_elements
from mmgroup.tests.test_gen_xi.test_gen_type4 import mul_v3
from mmgroup.tests.test_gen_xi.test_gen_type4 import mul_v2
from mmgroup.tests.test_gen_xi.test_gen_type4 import str_v3
from mmgroup.tests.test_gen_xi.test_gen_type4 import weight_v3
from mmgroup.tests.test_gen_xi.test_gen_type4 import chisquare_v3



#####################################################################
# Creating test vectors
#####################################################################

# Define STD_TYPE3 = -(5,1,1,...,1) to be the standard type-3 
# vector in theLeech lattice.

# Here is STD_TYPE3 (mod 3) in Leech lattice mod 3 encoding
STD_TYPE3_MOD3 = 0xfffffe000001
# Here is STD_TYPE3 (mod 2) in Leech lattice encoding
STD_TYPE3_MOD2 = 0x800800







#####################################################################
# Test mapping of type 3 Leech vectors modulo 3 to vectors modulo 2
#####################################################################


def v3_to_v2(v3):
    """Map type-3 vector from Leech lattice mod 3 to Leech lattice mod 2

    Here parameter ``v3`` is a type-3 vector in the Leech lattice mod 3 
    in Leech lattice mod 3 encoding. 

    The function returns a type-3 vector in the Leech lattice mod 2
    corresponding to ``v3`` in Leech lattice encoding.

    The result is unique. The function returns 0 if ``v3`` is not of
    type 3 in the Leech lattice mod 3
    
    Tis function is a wrapper for the C function ``gen_leech3to2_type3`` 
    in file ``gen_leech3.c``. 
    """
    result = gen_leech3to2_type3(v3)
    assert result != 0, (str_v3(v3), weight_v3(v3), hex(result))
    return result




@pytest.mark.gen_xi
def test_type3(verbose = 0):
    r"""Test conversion of type-3 vectors 

    Let STD_TYPE3 be the type-3 vector in the Leech lattice defined
    above. Let STD_TYPE3_MOD2 and STD_TYPE3_MOD3 be the images of
    STD_TYPE3 in the Leech lattice mod 2 and mod 3, respectively.

    For a set of elements g of the group ``G_x_0`` we convert 
    STD_TYPE3_MOD3  * g to  a vector v2 in the Leech lattice mod 2  
    with function ``v3_to_v2`` and we check that the result is equal 
    to  STD_TYPE3_MOD2 * g. We use function ``create_test_elements`` 
    for generating the elements g. 
    """
    weights = defaultdict(int)
    for ntest, data in enumerate(create_test_elements()):
        g = MM0(data) 
        v3_st = STD_TYPE3_MOD3 
        v2_st = STD_TYPE3_MOD2       
        if verbose:
            print("\nTEST %s" % (ntest+1))
            print("v3_start = " , str_v3(v3_st))
            print("g =", g)
        v3 = mul_v3(v3_st, g)
        w = weight_v3(v3)  
        weights[w] += 1        
        v2_ref = mul_v2(v2_st, g) & 0xffffff
        v2 = v3_to_v2(v3) 
        ok = v2 == v2_ref 
        #if  weights[w] <= 20:
        #     assert  v2 == py_gen_leech3to2_type3(v3)        
        if verbose or not ok:
            if not verbose:
                print("\nTEST %s" % (ntest+1))
                print("v3_start = " , str_v3(v3st))
                print("g =", g)
            print("v3 = v3_st*g =",str_v3(v3))
            print("weight =",w)
            print("v2 obtained= ", hex(v2))
            print("v2 expected= ", hex(v2_ref))
            if not ok:
                ERR = "Error in opation mod 3"
                raise ValueError(ERR)
    print("weights =", dict(weights))
    assert set(weights.keys()) == set([9, 12, 21, 24])
    
    
    
#####################################################################
# Chisquare test of random type 3 Leech vectors modulo 3
#####################################################################



"""Return binommial coeffient n choose k"""
def binom(n, k):
    return int(scipy.special.binom(n, k) + 0.1)

# The dictionary contains the number DATA_GEOMETRY[w] of type-3 
# vectors in the Leech lattice modulo 3 of weight w. 
# This table is obtained from :cite:`Iva99`, Lemma 4.4.1
DATA_GEOMETRY = {
  24: 24 * 2**12, 
   9: 759 * 16 * 2**8,
  21: binom(24,3) * 2**12,
  12: 2576 * 2**11,
}  
 
# Number of type-3 vectors in the Leech lattice
NUM_LEECH_TYPE3 = 2**24 - 2**12
# .. must be equal to the sum of the values in DATA_GEOMETRY
assert sum(DATA_GEOMETRY.values()) ==  NUM_LEECH_TYPE3
# Inverse of the number of vectors in Leech lattice mod 3
I_NUMV3 = 3.0**(-24)

# Assemble a dictionary DICT_P for a chisquare test.
# That dictionary maps a bit weight w to itself, if that bit weight
# occurs as a weight of a type-3 vector in the Leech lattice mod 3 
# with sufficiently high probability. It maps w to 1 if w occurs as
# such a weight with low probability.
# Let P be a dictionary that maps y the probability that a random 
# vector v in the Leech lattice mod 3 has type 3 and that 
# DICT_P[weight(v)] is equal to y. Let P[0] be the probablity that
# such a random vector y is not of type 3.
BLOCKSIZE = 1000000  # Minimum of type-3 vectors needed for test 
DICT_P = defaultdict(int)
MIN_P = 1.0 / 40000
P = defaultdict(float)
for w, num in DATA_GEOMETRY.items():
    p = num * I_NUMV3
    assert 0 <= p < 1
    DICT_P[w] = 1 if p < MIN_P else w
    P[DICT_P[w]] += p
P[0] = 1.0 - NUM_LEECH_TYPE3 * I_NUMV3
DICT_P[0] = 0

P_MIN = min([x for x in P.values() if x > 0])

RANDMOD3 = [0, 1, 0x1000000]
def rand_v3():
    """Return a random vector in the space GF(3)^{24}.

    This random vector is encoded in **Leech lattice mod 3 encoding**.
    """
    randomlist = choices(RANDMOD3, k=24)
    return sum((x << i for i, x in enumerate(randomlist)))
    
def rand_v3_dict(n = BLOCKSIZE):
    """Create n random vectors and group the according to their weight

    We create n random vectors in GF(3)^{24}. We return a dictionary
    ``d`` with the following entries:

    d[0] counts the vectors not of type 3 in the Leech lattice mod 3.
    d[w] counts the vectors v of type 3 such that  DICT_P[weight(v)]
    is equal to w.

    Then the value d[i] should be about P[i] / n.    
    """
    d = defaultdict(int)
    for i in range(n):
        v3 = rand_v3()
        v2 = gen_leech3to2_type3(v3) 
        if (v2 == 0):
            d[0] += 1
        else: 
            w = mat24_bw24((v3 | (v3 >> 24)) & 0xffffff)
            d[DICT_P[w]] += 1
    return d






@pytest.mark.slow
@pytest.mark.very_slow
@pytest.mark.gen_xi
def test_chisq_type3(verbose = 0):
    """Test distribution of weights of type-t vectors

    The function creates a list of n random vectors in GF(3)^{24}
    and groups them with respect to the property of having type 3
    in the Leech lattice and with respect to the weight as 
    described in function ``rand_v3_dict``.

    A chisquare test fails if the p-value is less than 0.01.
    We perform at most 4 chisquare test and raise ValueError
    if all of them fail.
    """
    p_min = 0.01
    print("Check distribution of type-3 vectors mod 3") 
    for i in range(4):
        d = rand_v3_dict()  
        chisq, p =  chisquare_v3(d, P)
        if verbose or i or p < p_min:
            print("Chisq = %.3f, p = %.4f" % (chisq, p))
        if p >= p_min: return
    raise ValueError("Chisquare test failed") 






"""The module contains class ``BitMatrixMulTransp``.

Class ``BitMatrixMulTransp`` contains a directive that generates
code to multiply a bit vector with the transposed matrix of a
fixed bit matrix. Depending on the size of the matrix and the
bit length of the underlying integer type, several bit vectors
can be multiplied with the same matrix simultaneously.

The C function ``mat24_autpl_set_qform`` uses that matrix 
multiplication, see section :ref:`implement-autpl-label`
in the *Guide for developers* for background.
"""

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

from random import randint
import types
import sys
import re
import os
from operator import __or__, __xor__

from mmgroup.generate_c import UserDirective, UserFormat


class BitMatrixMulTransp(object):
    """Generate code for multipling a bit vector with a bit matrix

    The generated code multiplies several bit vectors ``v_i`` by the 
    transposed of some corresponding bit matrices ``m_i``.

    Bit matrices ``m_i`` and bit vectors ``v_i`` have fixed size. 
    Vector ``v_i`` is ``lv`` bits long, where ``lv`` must be a power 
    of  two  or three times a power of two. Matrix ``m_i`` has ``lm`` 
    rows and ``lv`` columns, with ``lm <= lv``. We will actually 
    multiply  ``n`` vectors ``v_i`` by ``n`` matrices ``m_i``. Bit 
    ``v_i[k]`` is stored in bit ``i *lv + k`` of the integer ``v`` 
    and bit ``m_i[j, k]`` is stored in bit ``i *lv + k`` of the 
    integer``m[j]`` in the integer array ``m``. ``n * lv  <= 64`` 
    must hold. The result is stored in the variable ``v`` in the 
    same format as the input ``v``.

    The sizes ``lv, lm, n`` must be set with method ``set_matrix``.
    Variable names ``m`` and ``v`` must be set with method 
    ``set_names``.
     
    ``v`` should be an integer and ``m`` should be an array of 
    integers of type ``uint_fast64``. In case ``n * lv <= 32`` 
    that type may also be ``uint_fast32``.  
    """ 

    def set_matrix(self, lv, lm, n=1):
        """Set lv, lm, n  as in the description of the class"""
        assert 0 < lm <= lv <= lv * n <= 64, (lm, n, lv)
        self.lv, self.lm, self.n = lv, lm, n
        self.wordlength = 32 if n * lv <= 32 else 64
        self.int_fmt = "0x%xUL" if self.wordlength == 32 else "0x%xULL"
        self.int_type = "uint_fast%d_t" % self.wordlength
        self.masks = {}
        self.all_mask = (1 << (n * lv)) - 1
        self.nops = 0
        self.temp_name = "t_"
        self.n_temps = self.max_temps = 0

    def set_names(self, v, m, temp = "t"):
        """Set names of bit vector ``v`` and bit matrix ``m``

        Optionally, a prefix for temporary variables may be set.
        """
        self.v, self.m, self.temp =  v, m, temp
        self.n_temps = 0
        self.nops = 0

    def mask(self, space, width, pos):
        """Return a certain bit mask as a C literal

        In the returned bit mask we set the bits at positions

        ``k * space + pos * width + j,   0 <= j < width``,

        for all ``k >= 0`` such that the bit position is less than
        ``n * lv``, with ``n, lv`` as given by method ``set_matrix``.

        ``space`` must be divisible by ``lv`` and ``width`` must be
        divisible by ``space``.
 
        Since the same mask is used frequently, we use a dictionary
        for storing the masks already created. 
        """
        try:
            return self.masks[(space, width, pos)]
        except KeyError:
            assert self.lv % space == space % width == 0
            m = self.all_mask // ((1 << space) - 1) 
            m *= (1 << width) - 1
            m = m << (pos * width)
            s = self.masks[(space, width, pos)] = self.int_fmt % m
            return s
 
    def var(self, i):
        """Return the name of the i-th temporary variable"""
        return  self.temp_name + str(i)

    def new_var(self):
        """Return the name of the next unused temporary variable

        The returned variable is marked as used.
        """
        v = self.var(self.n_temps)
        self.n_temps += 1
        self.max_temps = max(self.n_temps, self.max_temps)
        return v

    def release_vars(self, n_vars = 1):
        """Mark temporary variables as unused

        We mark the ``n_vars`` last recently used variables as unused.
        ``n_vars`` defaults to 1.
        """
        self.n_temps -= n_vars
   
    def gen_start(self, i):
        """Generate code to process ``m[i]`` 

        The function returns a pair ``(var, code)``. Here ``var``
        is a new temporary variable containing the result and ``code``
        is the C code generated for computing that result.

        Bit ``[k]`` of the result ``var`` is equal to``v[k] & m[i,k] 
        for all relevant indices ``k``.  ``var`` is None if the result 
        is always zero. This happens if ``i >= self.lm``.
        """
        if i >= self.lm:   
            return None, ""
        tmp = self.new_var()
        c = "{tmp} =  {m}[{i}] & {v};\n".format( tmp = tmp,
             m = self.m, i = str(i), v = self.v )
        self.n_ops += 1
        return tmp, c

    def shl(self, var, shift):
        """Return the C expression ``(var << shift)``

        ``var`` must be the name of a variable and ``shift`` must
        be an integer corresponding to a fixed shift factor.  In case 
        ``sh < 0`` we generate the expression ``(var >> -shift)``.
        """
        if shift == 0:
            return var
        elif shift > 0:
            self.n_ops += 1
            return "(%s << %d)" % (var, shift)
        else:
            self.n_ops += 1
            return "(%s >> %d)" % (var, -shift)
     

    def merge(self, var, width, blocks, pos):
        """Generate code for XORing blocks of variable ``var``

        For the integer variable ``var`` the generated code computes
        the XOR sum ``x`` sum of the values ``var >> (i * width)``,
        ``0 <= i < blocks``. Then ``x`` is masked with the mask
        ``self.mask(blocks * width, width, 0)`` and shifted
        left by ``pos * blocks`` bits. The (masked and shifted) 
        result ``x`` is stored in variable ``var``.
        """
        assert blocks > 0
        if not var or blocks == 1:
            return ""
        parts = [self.shl(var, width * (pos-i)) for i in range(blocks)]
        self.n_ops += blocks
        mask = self.mask(width * blocks, width, pos)
        return "%s = (%s) & %s;\n" % (var, " ^ ".join(parts), mask)
  

    def gen_step(self, st, width):
        """Generate code to process ``m[st],...,m[st + width - 1]`` 

        The function returns a pair ``(var, code)``. Here ``var``
        is a new temporary variable containing the result and ``code``
        is the C code generated for computing that result.

        Bit ``[k*width + j]`` of the result ``var`` is equal to

        ``sum(v[l] & m[st+j, l] for l in range(k*width, (k+1)*width))``

        for ``0 <= j < width``. Here ``width`` must be a factor of
        the bit length of ``v``  and it should be a power of two
        or three times a power of two.

        The function recursively divides the blocks of length 
        ``width`` into two or three block of equal length
        and processes the parts of block.
       
        So in case ``width = 1`` the result is just ``v & m[st]``, 
        which is the content of the variable returned by of member 
        function ``self.gen_start(st).

        In case ``width = lv, st = 0`` the content of bit ``[i*lv + j]``
        of the the  variable ``var``  returned by this function with
        is eqaual to

        ``sum(v[l] & m[j, l] for l in range(i*lv, (i+1)*lv)``

        which is just bit ``j`` of the requested matrix product
        of the bit vector ``v_i`` with the transposed  matrix of 
        the bit matrix ``m_i``. 

        The function returns the pair ``(None,"")`` if the result
        is equal to zero. This happens in case ``st >= self.lm``

        Though the code generation process is recursive, the generated 
        code is very efficient and, obvioulsy, not recursive.
        """
        if st >= self.lm:   
            return None, ""
        if width == 1:
            return self.gen_start(st)
        if width & (width - 1) == 0:
            blocks = 2
        elif width % 3 == 0:
            blocks = 3
        else:
            err = "Illegal vector length in class BitMatrixMulTransp"
            raise ValueError(err)
        width = width // blocks
        var, c = self.gen_step(st, width) 
        m = self.merge(var, width, blocks, 0)
        code = c + m
        for j in range(1, blocks):
            var1, c = self.gen_step(st + j * width, width)
            if var1:
                m = self.merge(var1, width, blocks, j) 
                code += c + m
                code += "%s |= %s;\n" % (var, var1)
                self.n_ops += 1
                self.release_vars(1)
        return var, code  
            
    def define_temps(self):
        """Generate code for defining temporary variables"""
        return "%s %s;\n" % (self.int_type, 
             ", ".join([self.var(i) for i in range(self.max_temps)]))

    def generate(self):
        """Generate code for doing the requested matrix multiplication"""
        self.n_ops = 0
        var, code = self.gen_step(0, self.lv)
        out = "// %d operations\n%s = %s;\n" % (self.n_ops, self.v, var)
        s = self.comment() + "{\n" + self.define_temps() + code + out + "}\n"
        return s

    def comment(self):
        s = """// Compute ``v_i = v_i * transpose(m_i), i = 0,...,{n1}``, ``v_i`` 
// a bit vector of length {lv}, ``m_i`` a {lm} times {lv}  bit 
// matrix. ``v_i`` is coded in the integer ``{name_v}``, ``m_i`` 
// is coded in the array  ``{name_m}`` of integers. Bits ``v_i[k]`` 
// and ``m_i[j][k]`` are coded in bits ``{lv}*i + k`` of the 
// variables  ``{name_v}`` and ``{name_m}[j]``, ``j = 0,...,{lm1}``, 
// respectively. 
""".format(n1 = self.n-1, lv = self.lv, lm = self.lm, lv1 = self.lv - 1, 
        lm1 = self.lm - 1, name_v =  self.v, name_m = self.m ) 
        return s
        
    def n_operations(self):
        """Return number of Boolean operations required"""
        self.gen_step()
        return self.n_ops

    def generate_c_mul_transp(self, v, m, lv, lm, n=1):
        """Generate code for multiplying bit vector with matrix

        ``v`` is the name of an integer encoding ``n`` bit vectors
        of length ``lv``. ``m`` is the name of an array of integers 
        encoding ``n`` bit matrices of shape ``lm`` times ``lv``.

        More details of the computation are given in the description
        of the class. The computedbit vectors are encoded in ``v``.
        """
        n = max(1, int(n) if n else 0)
        self.set_matrix(int(lv), int(lm), n)
        self.set_names(str(v), str(m)) 
        c = self.generate()
        #print(b.n_ops, "operations for generate_c_mul_transp", lv, lm, n)
        return c

    def parity(self, v):
        """Return bit parity of the unsigned 64-bit integer ``v``"""
        v ^= v >> 32; v ^= v >> 16; v ^= v >> 8; v ^= v >> 4
        return (0x6996 >> (v & 0x0f)) & 1

    def compute(self, v, m):
        """Compute requested matrix product of ``v`` and ``transpose(m)``

        Here ``v`` is an integer and ``m`` is a list of integers. The
        matrix product given by ``lv, lm, n`` in method ``set_matrix``
        (as described in the header of the class) is computed and
        returned as an integer.
        """
        res = 0
        rows = m[:self.lm]
        for i in range(self.n):
            sh = i * self.lv
            v1 = sum(self.parity((v & r) >> sh) << j 
                for j, r in enumerate(rows))
            res  |= v1 << sh 
        return res
    
    def extract(self, var, start, length, dest = 0):
        """Return C expression containing some bits of ``var``
        
        Here ``var`` is an integer variabbe. The function extracts
        bits ``start, ..., start + length - 1`` variable and returns
        an integer expression where these bits are copied to bits
        ``start, ..., start + length - 1``. The other bits  of the  
        returned expression are cleared.
        """
        start, length = int(start), int(length)
        dest = int(dest) if dest else 0
        v = self.shl(var, dest - start)
        mask = ((1 << length) - 1) << dest
        fmt = "0x%xUL" if dest + length <= 32 else "0x%xULL"
        return "(%s & %s)" % (v, fmt % mask)

    def tables(self):
        """This class generates no tables"""
        return {
           "BITV_EXTRACT" : UserFormat(self.extract, "sii")
        }

    def directives(self):
        """Code generation directive has name "BITVMULTRANSP". """
        return {
           "BITVMULTRANSP" : UserDirective(self.generate_c_mul_transp, "ss"),
        }




from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import sys
import os
import numpy as np
from random import randint

import pytest


from mmgroup.mm import mm_sub_get_table_xi
from mmgroup.tests.groups.mgroup_n import MGroupNWord
from mmgroup.tests.spaces.sparse_mm_space import SparseMmV


space = SparseMmV(3)
group = MGroupNWord

 

def unit_vector_to_tuple(v):
    tuples = v.as_tuples()
    assert len(tuples) == 1
    t = tuples[0]
    assert t[0] == 1
    return t[1:]


def tuple_to_xi_tuple(t):
    tag, i, j = t
    if tag == 'B':
        return "BC", 32 * i + j
    if tag == 'C':
        return "BC", 32 * (i + 24) + j
    if tag == 'T':
        if 0 <= i < 15:
            return "BC", 32 * 48 + i *64 + j
        if 15 <= i < 375:
            return "T0", 64 * (i - 15) + j
        return "T1", 64 * (i - 375) + j
    if tag == 'X':
        if i < 1024:
            return "X0", 32 * i + j
        return "X1", 32 * (i - 1024) + j


def xi_tuple_to_tuple(t):
    tag, i = t
    if tag == "BC":
        if i >= 48 * 32:
            i1 = i - 48 * 32
            return "T", i1 >> 6, i1 & 63
        if i > 24 * 32:
            i1 = i - 24 * 32
            return "C", i1 >> 5, i1 & 31
        return "B", i >> 5, i & 31
    if tag == "T0":
        return "T", (i >> 6) + 15, i & 63
    if tag == "T1":
        return "T", (i >> 6) + 375, i & 63
    if tag == "X0":
        return "X", (i >> 5), i & 31
    if tag == "X1":
        return "X", (i >> 5) + 1024, i & 31


def op_xi_tuple(v, e, verbose =  0):
    from mmgroup.dev.mm_basics.mm_tables_xi import MM_TablesXi
    tables_xi = MM_TablesXi()
    e1 = e - 1
    table_index = None
    for i, source_tag in tables_xi.SOURCE_TAGS.items():
        if source_tag[e1] == v[0]:
            table_index = i
    dest_tag = tables_xi.DEST_TAGS[table_index][e1]
    dest_shape = tables_xi.DEST_SHAPES[table_index]
    source_shape = tables_xi.SOURCE_SHAPES[table_index]
    block, entry = divmod(v[1], source_shape[1] * 32)
    section_len = dest_shape[1] * dest_shape[2]
    section = block * section_len
    ti1 = table_index - 1
    if verbose > 1:
        print("xi_tuple table", e1, ti1,  "; ", block, entry, source_shape)
        print("destination",  dest_tag , dest_shape) 
    ref_table = tables_xi.PERM_TABLES[e, table_index]
    ref_part = ref_table[section : section + section_len] 
    index = section + int(np.nonzero((ref_part & 0x7fff) == entry)[0][0])
    assert mm_sub_get_table_xi(e1, ti1, index, 0) & 0x7fff == entry
    if verbose > 1:
        print("dest tag", dest_tag, ", index", index)
    q, r = divmod(index, dest_shape[2])
    dest_index = 32 * q + r
    dest_sign = (mm_sub_get_table_xi(e1, ti1, q, 1) >> r) & 1
    ref_sign_table = tables_xi.SIGN_TABLES[e, table_index]
    assert dest_sign == (ref_sign_table[q] >> r) & 1
    return dest_sign, (dest_tag, dest_index)



def op_xi_table(v, exp, verbose):
    assert 1 <= exp <= 2
    exp1 = exp
    v_tuple = unit_vector_to_tuple(v)
    v_xi = tuple_to_xi_tuple(v_tuple)
    if verbose:
        print("xi tuple =", v_xi)
    sign, w_xi = op_xi_tuple(v_xi, exp1, verbose)
    w_tuple = xi_tuple_to_tuple(w_xi)
    if verbose:
       print("result tuple =", sign, w_tuple, ", xi_tuple =", w_xi)
    return (-1)**sign * space(*w_tuple)


def op_xi_ref(v, exp):
    g = group("l", exp)
    return v * g 


#####################################################################
# Tests
#####################################################################

def xi_testcases(n_cases = 50):
    test_cases = [
        (("B", 2, 1), 1),
        (("B", 12, 1), 1),
        (("C", 2, 1), 1),
        (("C", 12, 1), 1),
        (("T", 14, 1), 1),
        (("T", 21, 13), 2),
        (("T", 16, 1), 1),
        (("T", 18, 0), 1),
        (("T", 38, 10), 1),
        (("T", 376, 60), 1),
        (("T", 437, 63), 1),
        (("X", 0x400, 10), 2),
        (("X", 0x444, 10), 2),
    ]
    for v0, exp in test_cases:
        yield space(*v0), exp
    for i in range(n_cases // 2):
        i, j, e = randint(0,14), randint(0, 63), randint(1,2)
        yield space("T", i, j), e
    tags = "BCTX" 
    for v_tag in tags:  
        for i in range(n_cases):
            v = space(v_tag, "r")
            for exp in (1,2):
                yield v, exp
 

def one_test_op_xi(v, exp, verbose = 1):
    w_ref = op_xi_ref(v, exp)
    if verbose:
        print("\nCompute ", v, "* t**", exp, "=", w_ref)
    w = op_xi_table(v, exp, verbose)
    ok = w == w_ref
    if verbose or not ok:
        print(v, "* xi**", exp, "=")
        print(w)
        if not ok:
            print("expected:")
            print(w_ref)
            print("product with inverse:")
            print(op_xi_ref(v, 3-exp))
            raise ValueError("Monomial operation of xi failed")
  


@pytest.mark.mm_op
def test_op_xi(n_cases = 150, verbose = 0):
    print("Testing monomial operation of xi")
    i = 0
    for args in xi_testcases(n_cases):
        one_test_op_xi(*args, verbose = verbose)
        i += 1
    print("Test passed")








from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals


import pytest

from random import randint

from mmgroup import mat24
mat24fast = mat24


from mmgroup.bitfunctions import bit_mat_mul, v2
from mmgroup.dev.mat24.mat24theta import theta_to_basis_vector


from mmgroup.dev.mat24.mat24_ref import Mat24



#####################################################################
# Test operation of the group Mat24 on Golay code and cocode
#####################################################################



def Mat24_testcases():
    """yields pairs (v,k) for testing the implementation of Mat24

    v is a Golay code vector. k is an integer representing an element 
    of the Mathieu group Mat24. You may use the Mat24 method 
    m24num_to_perm to convert k to a permutation.

    """
    testdata = [
       (0,  0),
       (0xdb1235, 115873693 ),
       (0xffffff, 244823040-1 ),
       (0,  244823040-1),
    ]
    for t in testdata: yield t
    for i in range(300):
        yield randint(0, 0xffffff), randint(0,244823040-1)


@pytest.mark.mat24
@pytest.mark.parametrize("gc, ref, small", [(mat24fast, Mat24, True)])
def test_Mat24(gc, ref, small):
    """test implementation of the Mathieu group Mat24"""
    for i, (v, k) in enumerate( Mat24_testcases()):
        p =  gc.m24num_to_perm(k)
        if ref and i < 20:
            p_ref = ref.m24num_to_perm(k)
            assert p == p_ref, (p, p_ref, k)
        assert sum(p) == 12*23
        v1 = gc.op_vect_perm(v, p)
        assert gc.vect_type(v) == gc.vect_type(v1), (
            map(hex,[v,k,v1]), i, gc.vect_type(v),gc.vect_type(v1))
        k1 = gc.perm_to_m24num(p)
        if ref and i < 20:
            k1_ref = ref.perm_to_m24num(p)
            assert k1 == k1_ref, (i, k1, k1_ref)
        assert k == k1,  (map(hex, [v, k, k1]),i)
        gc.perm_check(p)
        #if (i & 0x3ff) == 0x3ff: print( '.', end ="" )

        vc = randint(0,0xfff)
        m = gc.perm_to_matrix(p)
        vc1 = gc.vect_to_gcode(
               gc.op_vect_perm(gc.gcode_to_vect(vc), p)   )
        vc2 = gc.op_gcode_matrix(vc, m)
        vc2a = gc.op_gcode_perm(vc, p)
        assert vc1 == vc2, (list(map(hex,[vc,vc1,vc2])),p,list(map(hex,m)))
        assert vc1 == vc2a

        vc3 = gc.op_cocode_perm(vc, p)
        vc4 = gc.op_vect_perm(gc.cocode_to_vect(vc), p)
        vc4a = gc.vect_to_cocode(vc4)
        assert vc3 == vc4a, (hex(vc3), hex(vc4), hex(vc4a), p)

        p1 = gc.matrix_to_perm(m)
        assert p == p1

        if small and i > 100:
            break
        if ref and i < 100:
            assert p ==  ref.m24num_to_perm(k), (hex(k),p,ref.m24num_to_perm(k))
            assert k1 == ref.perm_to_m24num(p)
            assert vc2 == ref.op_gcode_matrix(vc, m)
            p1 = [p[1], p[0]] + p[2:]
            with  pytest.raises(ValueError):
                gc.perm_check(p1)
    print( "test of the Mathieu group operation passed" )
    



#####################################################################
# Test group operation of the group Mat24 
#####################################################################


def Mat24_group_testcases():
    """yields pairs (v,k1, k2) for testing the group operation of Mat24

    v is a Golay code vector. k1, k2 are integers representing an element 
    of the Mathieu group Mat24. You may use the Mat24 method 
    Mat24_int_to_perm to convert k1, k2 to a permutation.

    """
    testdata = [
    ]
    for t in testdata: yield t
    mmax = 244823040-1
    for i in range(40):
        yield randint(0, 0xffffff), randint(0,mmax), randint(0,mmax)



@pytest.mark.mat24
@pytest.mark.parametrize("gc, ref, small", [(mat24fast, Mat24, True)])
def test_group_Mat24(gc, ref, small):
    """test implementation of the Mathieu group operations in Mat24"""
    if gc is None: gc = Mat24
    for i, (v, k1, k2) in enumerate( Mat24_group_testcases()):
        p1 =  gc.m24num_to_perm(k1)
        p2 =  gc.m24num_to_perm(k2)
        p12 = gc.mul_perm(p1, p2)
        v1 =  gc.op_vect_perm(v, p12)
        v2 =  gc.op_vect_perm(gc.op_vect_perm(v, p1),p2)
        assert v1 == v2
        p2i = gc.inv_perm(p2)
        p1a = gc.mul_perm(p12, p2i)
        assert p1 == p1a
    print( "test of the Mathieu group implementation passed" )


 
#####################################################################
# Test operation of the group Aut_PLoop on the Parker loop 
#####################################################################

AUT_PL_SMALL = 20

def Aut_Pl_testcases(gc, small = False):
    """yield tuples (c1, p1, c2, p2, v1, v2) for testing Automorphisms of Pl

    (c1, p1) and (c2, p2) are pairs of a cocode element and an integer
    representation to form an automorphism of the Parker loop.
    v1 and v2 are elements of the Parker loop given as 13-bit integers   
    """
    Id = 0   # number of identity permutation 
    testcases = [
        (0, Id, 0, Id, 0x111, 0x222),
        (0x55, 567234, 0x8a5, None, 0x111, 0x222),
        (0x356, 567234, 1, None, 0x876, 0x9ab),
        (0, 1, 0, Id, 0x111, 0x222),
    ]
    for t in testcases:
        yield t
    mmax = gc.MAT24_ORDER-1
    def rand_aut():
        return randint(0, 0xfff), randint(0, mmax)
    def rand_v1v2():
        return randint(0, 0x1fff), randint(0, 0x1fff)
    for i in range(AUT_PL_SMALL if small else 200):
        yield  rand_aut() + rand_aut() +  rand_v1v2()
  

@pytest.mark.mat24
@pytest.mark.parametrize("gc, ref, small", [(mat24fast, Mat24, False)])
def test_Aut_Pl(gc, ref, small):
    """test Automorphism operations of the parker loop"""
    def to_aut(c, p, check = True):
        p = gc.m24num_to_perm(0 if p is None else p)
        m = gc.perm_to_autpl(c, p)
        if check and ref:
             mref = ref.perm_to_autpl(c, p)
             assert m == mref, (map(hex,m), map(hex,mref))
        return m

    def autpl_op_tuple(tuple_, a, g):
        return tuple([g.op_ploop_autpl(v, a) for v in tuple_])

    for i, (c1, p1, c2, p2, v1, v2) in enumerate(Aut_Pl_testcases(gc, small)):
        p = gc.m24num_to_perm(p1)
        m = gc.perm_to_matrix(p)
        check = i < AUT_PL_SMALL
        a1, a2 = to_aut(c1, p1, check),  to_aut(c2, p2, check) 
   
        # test composition and decomposition
        assert c1 == gc.autpl_to_cocode(a1)
        assert p == gc.autpl_to_perm(a1)
        a1_0 = gc.perm_to_autpl(0, p)
        assert a1 == gc.mul_autpl(gc.cocode_to_autpl(c1), a1_0)

        # v1, v2 are Parker loop vectors
        # a1, a2 are Parker loop automorphisms
        # test (v1 * a1) * (v2 * a1) =  (v1 * v2) * a1 
        v3 = gc.mul_ploop(v1, v2)
        if check and ref:
            assert  v3 == ref.mul_ploop(v1, v2)
        v11, v12, v13 = autpl_op_tuple((v1, v2, v3), a1, gc)
        if check and ref:
            assert (v11,v12,v13) == autpl_op_tuple([v1,v2,v3], a1, ref)
        v13a = gc.mul_ploop(v11, v12)
        assert v13 == v13a, (i, map(hex,[v13, v13a]))

        # test (v * a1) * a2 = v * (a1 * a2) for v = v1, v2, v3
        v21, v22, v23 = autpl_op_tuple((v11, v12, v13), a2, gc)
        v23b = gc.mul_ploop(v21, v22)
        assert  v23 == v23b
        a12 = gc.mul_autpl(a1, a2)
        if ref and check:
            assert a12 == ref.mul_autpl(a1, a2)
        v21a, v22a, v23a = autpl_op_tuple((v1, v2, v3), a12, gc)
        assert (v21, v22, v23) == (v21a, v22a, v23a)

        # test inversion: check that (a1 * a2) * a2**(-1) == a2
        a2i = gc.inv_autpl(a2)
        a1n = gc.mul_autpl(a12, a2i)
        assert a1 == a1n, (map(hex,a1), map(hex,a1n))
        # test inversion function perm_to_autpl_inv()
        perm2 = gc.m24num_to_perm(0 if p2 is None else p2)
        perm2i, a2ia = gc.perm_to_iautpl(c2, perm2)
        assert a2i == a2ia 
        assert perm2i ==  gc.inv_perm(perm2)   
        
    print( "Test of Automorphisms of Parker Loop passed" )






r"""We describe the class ``MM_Op``.

Class ``MM_Op`` in module ``mmgroup.dev.mm_op.mm_op`` is a 
table-providing class for the C functions used by python extensions
``mmgroup.mm<p>``. Each of these extensions implements
the operation of monster group modulo a fixed number ``p``. 

Class ``MM_Op`` provides the same functionality as class ``MM_Const``
in module ``mmgroup.dev.mm_basics.mm_basics``. Since modulus ``p``
is fixed in an instance of class ``MM_Op``, all constants provided
by the base class ``MM_Basics`` of ``MM_Op`` and ``MM_Const`` (see 
table *Constants used for generating representations of the monster*)
have fixed values. So they are available as attributes of an
instance of class ``MM_Op`` and they may also be used by the code 
generator directly via string formatting.
 
The string formatting functions ``%{shl:expression,i}`` and 
``%{smask:value, fields, width}`` work as in class ``MM_Op``. In 
addition, parameter ``fields`` defaults to ``-1`` (i.e. the mask 
is set in all bit fields) and ``width`` defaults to the constant 
``FIELD_BITS``. i.e. the number of bits used to store an integer 
modulo ``p``. E.g. for ``p = 7`` we have ``FIELD_BITS = 4``, and 
``%{smask:1}`` evaluates to the ``64``-bit integer constant 
``0x1111111111111111``.

Class ``MM_Op`` also provides the following directives:

``MMV_ROTL src, count, dest``

  Rotate the bit fields of a variable of type ``uint_mmv_t``

  Here ``src`` is an integer of type ``uint_mmv_t`` which is 
  interpreted as an array of bit fields, where each bit field
  stores a number modulo ``p``. Then each bit field is rotated 
  left by ``count`` bits. This means that the numbers in all
  bit fields are multiplied by ``2**count``. ``count`` must be 
  an integer. It is reduced modulo the bit length of ``p``.
  So ``count`` may also be negative. The result of the 
  rotation is written to the variable ``dest`` which should
  also be of type ``uint_mmv_t``. ``dest`` defaults to ``src``.

``MMV_UINT_SPREAD src, dest, value``
  
  Spread bits of an integer ``src`` to the bit fields of ``dest``.
  Here ``src`` may be any integer variable and ``dest`` should
  be a variable of type ``uint_mmv_t``. If bit ``i`` of the
  integer ``src`` is set then the ``i``-th bit field of variable
  ``dest`` is set to ``value``; otherwise it is set to zero.
  ``value`` must be an integer with ``0 <= value <= p``; 
  default is ``p``.

"""

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

import sys
import os
import warnings
from numbers import Integral
import re

from mmgroup.generate_c import UserDirective, UserFormat

from mmgroup.dev.mm_basics.mm_basics import INT_BITS, c_hex
from mmgroup.dev.mm_basics.mm_basics import smask, smask_default
from mmgroup.dev.mm_basics.mm_basics import MM_Basics


########################################################################
# Class MM_Op
########################################################################

class MM_Op(MM_Basics):
    r"""Supports basic operations on integers of type ``uint_mmv_t``.

    Here an integer of type ``uint_mmv_t`` is interpreted as an array 
    of bit fields, where each bit field stores an integer modulo ``p``
    for a fixed number ``p``. This class is similar to class 
    ``MM_Const`` in module ``mmgroup.dev.mm_basics.mm_basics``,
    where the modulus ``p`` may be variable.

    Usage of this class is documented in the module documentation.

    :param p: 

       This is the modulus ``3 <= p < 256``. ``p + 1`` must be a 
       power of two.

    :type p: int

    """
    m_psum = re.compile(r"\s*(\w+)\s*([+-].*)?$")

    def __init__(self, p):
        """Create instance for modulus p with p + 1 a power of two"""
        self.p = p
        self.directives =  {
            "MMV_ROTL" : UserDirective(self.gen_mmv_rotl, "sis"),
            "MMV_UINT_SPREAD": UserDirective(self.gen_mmv_uint_spread, "ss"),
        }
        self.tables = self.sizes(p).copy()
 
    def __getattr__(self, attrib):
        return self.tables[attrib]

    def smask(self, value, fields = -1, width = None):
        """Return smask(value, fields, width)

        'fields' defaults to -1, 'width' defaults to self.FIELD_BITS.
        """
        return smask_default(self.FIELD_BITS, value, fields, width)

    def gen_mmv_rotl(self, src, count, dest = None):
        """Rotate 'src' left by count 'bits', store result to 'dest'

        This generatates code for computing 
        'dest' = 'src' * 2**k  (mod p)
        for any fixed integer k. 'src' and 'dest' of type uint_mmv_t.
        They are interpreted as arrays of bit fields. The operation 
        is done on all bit fields simultaneously.
        """
        if dest is None:
            dest = src
        count = int(count) % self.P_BITS
        if count == 0:
            return
        s = """// Put %{dest} = %{src} rotated left by %{c} for all fields
%{dest} = (((%{src}) << %{c}) & %{smask: (-1 << {c}) & P})
       | (((%{src}) >> %{int:P_BITS-c}) & %{smask:range(c)});
"""
        return self.snippet(s, src = src, dest = dest, c=count)

    def gen_mmv_uint_spread(self, src, dest, value = None):
        """Spread bits of the integer 'src' to field array 'dest'

        'dest' is of type uint_mmv_t and interpreted an array of 
        bit fields. bit field i of 'dest' is set to -1 (mod p) if 
        bit i of 'src' is one and to 0 otherwise.

        Bit field i of 'dest' is set to  'value' if bit i of 'src' is 
        one. 0 <= 'value' <= self.p must hold. 'value' dafaults to
        self.p
        """ 
        msize = 1 << (self.INT_BITS >> 1)

        if value is None:
            value = self.P
        assert  0 <= value <= self.P
        if value == 0: 
            return "{dest} = 0;\n".format(dest = dest)
        s = """// Spread bits 0,...,{i} of {src} to the ({j}-bit long) fields
// of {dest}. A field of {dest} is set to {value} if its 
// corresponding bit in input {src} is one and to 0 otherwise.
""". format(i = self.INT_FIELDS - 1, value = hex(value), j = self.FIELD_BITS,
                      dest = dest, src= src)
        a, b = self.INT_FIELDS >> 1, self.INT_BITS
        while a:
            m0 = self.hex(self.smask(range(a), -1, b))
            m1 = self.hex(self.smask(range(a, 2*a), -1, b))
            b >>= 1
            s0 = "{dest} = ({src} & {m0}) + (({src} & {m1}) << {sh});\n"
            if len(m0 + m1) >= 17:
                s0 = "\n    + ".join(s0.split("+"))
            s += s0.format(dest=dest, src=src, m0=m0, m1=m1, sh=b-a)
            a >>= 1
            src = dest
        s += "%s *= %s;\n"  % (dest, value)
        s += "// Bit spreading done.\n"
        return s




class Mockup_MM_Op(MM_Op):
    def __init__(self, p):
        super(Mockup_MM_Op, self).__init__(p)
        old_tables = self.tables
        self.tables = {}
        self.tables.update(old_tables)
        self.tables["GENERATE_CODE"] = False









from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals

import pytest

from mmgroup.dev.mm_basics.mm_basics import MM_Basics
from mmgroup import mm

@pytest.mark.mm_op
def test_mm_aux_mmv_size():
    for i in range(2,9):
        p = 2**i - 1
        res = mm.mm_aux_mmv_size(p)
        ref = MM_Basics.sizes(p)["MMV_INTS"]
        assert res == ref, (res, ref, p)
    print("Test of function mm.mm_aux_mmv_size() passed")





r"""The C code generator generates code from a source file automatically.


    Class ``TableGenerator`` can be used to generate  a 
    .c and a .h file from *source* file. Here such a source file usually
    has the extension .ske.
 
    Basic idea: We copy a *source* file to the .c file. The source 
    contains directives to enter precomputed tables, code  snippets 
    or constants into the .c file, and prototypes into the .h file.

    We copy the content of the *source* file directly to the .c
    file.   In the *source* file we also parse certain directives 
    and we replace them by automatically generated code snippets 
    in the .c file and in the .h file. 

    The arguments given to the constructor of class ``TableGenerator``
    specify the meaning of the directives.

    Overview
    --------   

    In the *source* file we parse directives of the shape::

      // %%<keyword>  <arg1>, <arg2>, ...

    Each directive executes certain function associated to the
    ``<keyword>``. There are built-in and user-defined directives.
    Built-in directives are listed in section *Built in directives*.
    User-defined directives are  passed to the constructor in a
    dictionary ``directives`` with entries::

            <keyword> : <directive_function>. 
    
    If a directive is found in the source file we call::

       <directive_function>(*<all evaluated arguments>), 

    and the string returned by that function is merged into the
    generated  .c  file. The arguments of a directive must be given 
    in python syntax and they  are evaluated to python objects using 
    some safe version of the python ``eval()`` function. Basically, 
    python identifiers in an argument are evaluated as follows:

    In the constructor, parameter ``tables`` is a dictionary of local
    variables, which will be passed to the python ``eval()`` function 
    for evaluating an argument. See section *Arguments of directives* for
    a more detailed description of this process.
      
    You may also consider class ``UserDirective`` 
    for constructing specific directive functions.


    Arguments of directives
    -----------------------
     
    The comma-separated list of arguments of a built-in or user-defined 
    directive or function must be a valid python expression. This 
    expression is evaluated by a safe version of the python ``eval()`` 
    function,  with the local variables given by the dictionary 
    ``names``. 
    Here ``names`` is  an updated version of the dictionary ``tables`` 
    of local variables  passed in the constructor of this class as
    described below. This means  that each identifier found in the list  
    of arguments is looked up in that dictionary, and that the identifier 
    is evaluated  to the corresponding value in that dictionary.

    The dynamic dictionary ``names`` is  initialized with dictionary  
    ``tables``. It may be updated  with the names of the tables generated 
    by the built-in directive ``TABLE``, as indicated in the
    description of the  built-in directive ``USE_TABLE``.

    There are a few more predefined entries in dictionary ``names``:

    ==========  ==============================================================
    name        value
    ==========  ==============================================================
    ``TABLES``  The original dictionary ``tables`` passed  in the  constructor
    ``NAMES``   updated version of dictionary ``tables`` as described above
    ==========  ==============================================================

    Evaluating a python expression with the standard function ``eval()`` 
    is known to be unsafe. For evaluating arguments we use a modified 
    version of the ``eval()`` function. That modified function does not
    allow assignment, and identifiers starting with an underscore ``_`` 
    are illegal. See function ``EvalNodeVisitor`` in module
    ``mmgroup.generate_c.generate_functions``. for details. 
    
    There is a limited number of built-in python functions (specified 
    in directory ``safe_locals`` in the same module) that are
    legal in python expressions for arguments::

       abs, int, len, max, min, str, pow, range, zip 

    An example with user-defined  directives is given in class 
    ``testcode.TestGen``. We recommend  to follow the conventions in that 
    class for passing user-defined directives to an instance of class 
    ``TableGenerator``.   


    Built-in directives
    -------------------

     * ``COMMENT``,  deprecated!
        Indicates that a comment follows. This should be used only
        for comments relevant for the user, not for internal details.
        No action in this class, but other functions may use this
        directive for generating user documentation.  

     * ``ELSE [IF]?`` <expression>
        *else* clause for IF directive, see ``IF``

     * ``END`` <directive>
        Closes a block that has been started by one of the directives
        ``FOR`` or ``IF``. A ``FOR`` block must be closed with ``END FOR``, 
        an ``IF`` block must be closed with ``END IF``, etc.

     * ``ERROR``  <message>
         TODO: yet to be implemented and documented!!!!


     * ``EXPORT``   (arguments see below)
        Then the following line with an appended semicolon is copied 
        into the generated .h file. This is useful for 
        creating prototypes.
        The ``EXPORT`` directive may precede a line with a ``USE_TABLE`` 
        directive.
        Then the table name captured by the next ``USE_TABLE``  statement 
        is  exported to the generated .h file. 
        The ``EXPORT`` directive may precede a line with a  ``DEFINE``  function.
        Then the #define statement generated by the next DEFINE function
        is written to the generated .h file.

        An optional argument ``p`` means that the exported line will also
        be included in a .pxd file, see method ``generate_pxd()``.

        An optional argument ``px`` or ``xp`` means that the same as 
        argument ``p``. Then in addition we also write a comment 
        ``# PYX <wrap pxi>`` into the .pxd file generated by method 
        ``generate_pxd()``. If function ``mmgroup.generate_c.pxd_to_pyx``
        reads such a comment in a .pxd file, it will write a simple wrapper
        for the exported function into a .pyx file.


     * ``EXPORT_KWD`` <keyword>
        The directive:: 

          // %% EXPORT_KWD  FOO_API

        places the keyword FOO_API before any function or variable 
        which is exported with the ``EXPORT`` or ``EXPORT_TABLE``
        directive.
        This is useful on a MS Windows system for creating a dll.
        Then the header should contain a definition similar to::

          #define FOO_API __declspec(dllexport)

        This declares these functions or variables to be exported by the
        dll. For background, see e.g. 
        https://gcc.gnu.org/wiki/Visibility.

     * ``EXPORT_TABLE`` (no arguments)
        Parses the next input line for the C-name of a table, and 
        exports this name into the generate .h header file.
        This is equivalent to two subsequent directives::

           EXPORT
           USE_TABLE

        There are, however, subtle differences, see directive 
        ``EXPORT_KWD``.

     * ``FOR``  <variable> ``in`` <parameter_list>
        Generate a sequence of blocks of statements, one block for 
        each parameter in the parameter_list. A sequence of arbitrary
        lines follows, terminated by an ``END FOR`` directive. The
        parameter list is evaluated to a python object which
        should be iterable.

        Nesting of FOR and similar directives is possible. 
  
        Examples::   

         // %%FOR x in range(1,3)
         i%{x} += 2*%{x};
         // %%END FOR

        evaluates to::

         i1 += 2*1;
         i2 += 2*2;
            

     * ``GEN`` <ch>
        Directs output to .c or to .h file or to both, depending on
        whether the letter c or h is present in the first argument.

     * ``IF`` <expression>
        Conditional code generation depending on the boolean value
        of <expression>.  Syntax is::

          // %%IF <expression>
          <statements>
          // %%ELSE IF <expression>
          <statements>
          // %%ELSE 
          <statements>
          // %%END IF

        with one or more optional ``ELSE IF`` clauses and at most one
        final ``ELSE`` clause. Nested ``IF`` and ``FOR`` blocks are
        possible.

     * ``JOIN`` <infix>, <suffix>    
        A ``FOR`` directive must follow a ``JOIN`` directive. The 
        expressions generated by the following ``FOR`` directive
        are joined with the <infix> (which is usually an
        operator), and the <suffix> is appended to the
        whole generated expression. E.g.::

          a += 
          // %%JOIN*  " +", ";"
          // %%FOR* i in range(1,4)
            %{int:2*i} * table[%{i}]
          // %%END FOR

        evaluates to::
 
          a += 
            2 * table[1] + 
            4 * table[2] + 
            6 * table[3];

     * ``PY_DOCSTR`` <string>, <format>
         Creates a comment from a python docstring <string> and  
         writes this comment into the C file.
         If <format> is set to 1, the string is formatted, see
         section *String formatting*. 
         if <string> does not evaluate to a python string then we
         also try to use <string>.__doc__ .         

     * ``PYX`` <string>
        Write a comment of shape::

            # PYX <string>

        into the .pxd file, when generating a .pxd file. This has
        no effect on the generated .c or .h file. It may be used for
        automatically generating a .pyx file from a .pxd file.

     * ``TABLE``  <table>, <format>
        Enters all entries of the python array <table> into the C code 
        as an array of constants with an optional <format>. Only the 
        data of the array are written. Feasible <format> strings are 
        given in function format_number().

     * ``USE_TABLE``   (no arguments)
        Parses the next input line for the C-name of a table and makes 
        that C name available for user-defined directives. 

        If a table is coded  with  the ``TABLE`` directive and
        preceded by a ``USE_TABLE`` or ``EXPORT_TABLE`` directive then
        the  name of that table will be added to the dictionary ``names``.
        The key with the python name of that table will have the
        C name of that table as its value.

        Consider the following example::

         1:    // %%USE_TABLE
         2:    int  C_table[]  = {
         3:       // %%TABLE python_table, uint32
         4:    } 

        Here ``'python_table'`` must be a key in the dictionary ``tables``
        passed to the constructor of class ``TableGenerator``. The
        value for that key should be a list of 32-bit integers.
        Then the directive in line 3 creates the list of integer
        constants given by the list ``tables[python_table]``.

        The ``USE_TABLE`` directive in line 1 has the following effect.
        In the sequel the entry ``names['python_table']`` has the value
        ``'C_table'``, which is the C name of the table that has been
        generated. So we may use that entry in subsequent directives
        or via string formatting with curly braces.  
      
        The entries of table ``python_table`` are still available in  
        the form ``python_table[i]`` as before. ``python_table[:]`` 
        evaluates to  the whole table as before.

     * ``WITH``  <variable> = <value>
        Temporarily assign  <value> to <variable>. A sequence of 
        arbitrary lines follows, terminated by an ``END WITH`` 
        directive. Both, <variable> and <value>, may be tuples of 
        equal length; a nested tuple <variable> is illegal. 
        The <variable>, or the variables contained in the tuple
        <variable>, are temporarily added to the dictionary ``names`` 
        as keys, to that they can be used via string formatting.

        Nesting of WITH and other directives is possible. 
  
        Example::   

          // %WITH square, cube = int((x+1)**2), int((x+1)**3)
          i += %{cube} + %{square} + %{x};
          // %%END WITH

        Assuming ``x`` has value ``9``, this evaluates to::

          i += 1000 + 100 + 9;
         
            



    String formatting
    -----------------

    A line ``l`` in the source file may contain a string ``s``  with 
    balanced curly  braces and preceded by a '%' character.
    Then the standard ``.format()`` method is applied  to  the string 
    ``s[1:]`` obtained from  ``s`` by dropping the initial '%' 
    character. The  dictionary ``names`` is given to the formatting
    method as given as the list of keyword arguments. 

    E.g. in case ``names['x'] = y``, the expressions ``%{x.z}`` and 
    ``%{x[z]}``  refer  to ``y.z`` and ``y[z]``. Of course, the objects 
    ``y.z`` or ``y[z]`` must exist and evaluate to strings which are 
    meaningful in the generated C file.

    For formatting, the dictionary ``names`` is an updated version of 
    the dictionary ``tables`` passed in the constructor, as described 
    in section *Arguments of directives*.
                 
    We consider this to be a reasonable compromise between the standard 
    C syntax and the very powerful pythonic ``.format`` method for 
    strings. 


    Using functions via string formatting
    -------------------------------------

    User-defined functions can be invoked inside C code via the the
    string formatting operator ``%{function_name:arg1,arg2,...}``.
    Then the function with name ``function_name`` is called with the
    given arguments, and the result of the function is substituted 
    in the C code. 

    Such a user-defined function may be created with class 
    ``UserFormat``. The simplest way  to create such a  user-defined 
    function from a function ``f`` is to create an entry::

        "function_name" : UserFormat(f)

    in the dictionary ``tables`` passed to the constructor. Then
    ``%{function_name:arg1,arg2,...}`` evaluates to 
    ``str(f(arg1,arg2,...))``.
    Here arguments ``arg1, arg2`` must be given in python syntax and 
    they are processed in the same way as the arguments of a directive 
    described in section *Arguments of directives*. Then you may also 
    write ``function_name(arg1,..)`` inside an argument of a directive 
    or  a user-defined function; this expression evaluates to 
    ``f(arg1,...)``.

    Class ``UserFormat`` also contains options to control the 
    conversion of the result of function f to a string.

    There are a few built-in string formatting functions, as given
    by dictionary ``built_in_formats`` in module 
    ``generate_functions``. 

    There are a few predefined functions for string formatting:

      * ``%{int:x}``    returns ``x`` as a decimal integer 
      * ``%{hex:x}``    returns ``x`` as a hexadecimal integer
      * ``%{len:l}``    returns length of list ``l`` as a decimal int     
      * ``%{str:x}``    returns ``str(x)``
      * ``%{join:s,l}`` returns ``s.join(l)`` for string ``s`` and list ``l``

    See dictionary ``built_in_formats`` in module 
    ``mmgroup.generate_c.generate_functions`` for details.

    So, assuming that keys ``'a'`` and ``'b'`` have values ``1`` and ``10`` 
    in the dictionary ``tables`` passed to the code generator, we may 
    write e.g.::

       y%{a} += %{int:3*b+1};

    in the source file. This evaluates to::

       y1 += 31;
        

    Quiet form of a directive
    -------------------------

    For directives such as ``IF``, ``FOR``, ... there are also quiet 
    variants ``IF*``, ``FOR*``, ..., which perform the same operations, 
    but without writing any any comments about the directive into the 
    output file. E.g. in::
         
        // %%FOR i in range(50)
        printf("%d\n", i);
        // %%IF* i % 7 == 0 or i % 10 == 7
        printf("seven\n");
        // %%END IF
        // %%END FOR

    you probably do not want any messy comments about the IF directive
    in your generated code.

    Historical remark
    -----------------

    In versions up to 0.4 string formatting operator was ``{xxx}``
    instead of ``%{xxx}``. This had the unpleasent effect that
    valid C expressions have a different meaning in the input and
    and the output of the code generation process. This situation
    became unbearable after intoducing doxygen for documentation
    of C code. Then expressions like ``\f$ x_{1,1} \f$`` could not
    be coded in a reasonable way the old version.

"""




"""Test lecical orderin of the Mathieu group M_24

The C function mat24_m24num_to_perm() in file mat24_functions.c 
converts a number 0 <= n < 244823040 to a permutation in the Mathieu 
group M_24. The C function mat24_perm_to_m24num() in the same file 
reverses that operation.

The numbering of the elements of M_24 provided by these functions is
lexicographic (based on the ordering of the permutations corresponding
to the elements of M_24). In this script we test if that ordering is
indeed lexicographic. Therefore we generate a large number of random
numbers  0 <= n < 244823040 and compute the permutations for these 
numbers. The we sort the list of the generated numbers and we check
that the list of permutations corrsponding to that sorted list of 
numbers is also sorted.

Furthermore, function  mat24_m24num_to_perm() has two reference 
implementations in python and function mat24_perm_to_m24num()
has one refernece implementation in python. For a small fraction 
of the test data we also check that the the (slow) reference 
implementions of the two C functions compute the same result as the
(fast) C functions.

""" 

from __future__ import absolute_import, division, print_function
from __future__ import  unicode_literals




from random import randint
import pytest

# Import the extension containing the C functions
from mmgroup import mat24
# Import two reference implementations for mat24_m24num_to_perm() 
from mmgroup.dev.mat24.mat24heptad import py_mat24_int_to_perm
from mmgroup.dev.mat24.mat24heptad import mat24_int_to_perm
# Import a reference implementation for mat24_perm_to_m24num()
from mmgroup.dev.mat24.mat24heptad import mat24_perm_to_int


MAT24_ORDER =  244823040 


def mat24lex_testcases(ntests = 1000):
    """Yield random test numbers 0 <= n < 244823040

    """
    # Yield a large number of random permutation numbers
    for  i in range(ntests):
        yield randint(0, MAT24_ORDER - 1)
    # Yield the lowest 10 permutation numbers
    yield from range(10)
    # Yield the highest 10 permutation numbers
    yield from range(MAT24_ORDER - 10, MAT24_ORDER)
    # Yield a random intveral of 100 adjacent numbers
    r = randint(0, MAT24_ORDER - 100)
    for i in range(r, r+100):
        yield i


@pytest.mark.mat24
def test_mat24lex(ntests = 5000, verbose = 0):
    lindex = []; ldata = []
    for i, n in enumerate(mat24lex_testcases(ntests)):
        # Do  additional checks for the first few samples
        test_ref = i < 100
        # Compute permutation p with number n using the cython wrapper
        # mat24.m24num_to_perm() for function mat24_m24num_to_perm()
        p = mat24.m24num_to_perm(n)
        # For the first few random test data do:
        if test_ref:
            # Check result p against two slow reference implementations
            p_py_ref = py_mat24_int_to_perm(n)
            assert p == p_py_ref, (n, p, p_py_ref)
            p_ref = mat24_int_to_perm(n)
            assert p == p_ref, (n, p, p_ref)
        if verbose:
            print(i, n, p[:10])
        # Append pair (n i) to the list 'lindex' and put ldata[i] = p
        lindex.append((n, i))
           ### p[2] = randint(0,23) # This would destroy the order
        ldata.append(p)
        # Compute number n1 of permutation p using the cython wrapper
        # mat24.perm_to_m24num() for function mat24_perm_to_m24num()
        n1 = mat24.perm_to_m24num(p)
        # Then n == n1 must hold
        assert n1 == n, (hex(n),hex(n1), p)
        # For the first few random test data do:
        if test_ref:
            # Check result n1 against a slow reference implementation
            n1_ref = mat24_perm_to_int(p)
            assert n1_ref == n, (hex(n),hex(n1_ref), p)

    # Sort the index list 'lindex' by permutation numbers.
    lindex.sort()
    # Let n0 be the first permutation number in 'lindex' and
    # let p0 be the corresponding permutation.
    n0, index0 = lindex[0]
    p0 = ldata[index0]
    # For all subsequent permutation numbers n1 in the list 'lindex'
    # let p1 be the corresponding permutation.
    for n1, index1 in lindex[1:]:
        p1 = ldata[index1]
        if n1 > n0:
            # Check that permutation p1 is greater the the previous
            # permutation p0 (assuming that n1 is greater than the
            # previous number n0).
            assert p1 > p0
        elif n1 == n0:
            # Birthday paradoxon: n1 == n0 may (and will) happen
            assert p1 == p0
        else:
            raise ValueError("Something is going wrong here")
        # Update 'previous' number n0 and permutation p0
        n0, p0 = n1, p1





r"""We describe an implementation of the subgroup :math:`N_{0}` of the monster group.


The subgroup :math:`N_{0}` of the monster group of structure 
:math:`2^{2+11+2\cdot 11}.(\mbox{Sym}_3 \times\mbox{M}_{24})` has been 
described in :cite:`Con85`. Theorem 5.1 in :cite:`Seysen20` reduces the 
group operation in :math:`N_{0}` to easy calculations in the Parker loop 
:math:`\mathcal{P}`, the Colay cocode :math:`\mathcal{C}^*`, and the group
:math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}` of standard automorphisms of
:math:`\mathcal{P}`. The loops :math:`\mathcal{P}`, :math:`\mathcal{C}^*`, 
and :math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}` are described in section
:ref:`basic_label`. Module ``mat24_functions.c`` provides the required
functions for computing in these loops.

Using the notation in section :ref:`mmgroup-label` we may describe an 
element of  :math:`N_{0}` as a product:

.. math::
    \tau^t y_f x_e x_\delta x_\pi \; ; \quad 0 \leq t < 3; \; 
    e, f \in   \mathcal{P}; \; \delta \in \mathcal{C}^*; \;
    \pi \in  {{\rm Aut}_{{\rm St}} \mathcal{P}} \; .


This representation is unique if we require :math:`f` to be in a
transversal of :math:`\mathcal{P} / Z(\mathcal{P})` and 
:math:`\pi` to be a standard representative in
:math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}` as described in section 
:ref:`aut_ploop_label`.

We store an element of :math:`N_{0}` an a quintuple
:math:`t, f, e, \delta, \pi` of five integers of type ``uint32_t``.
For :math:`f, e` we use the numbering in class ``PLoop`` in module
``mmgroup``; for :math:`\delta` we use the numbering in class 
``Cocode`` in module ``mmgroup``. Here :math:`\pi` refers to a standard 
representative in :math:`{{\rm Aut}_{{\rm St}} \mathcal{P}}`. These 
standard representatives correspond to the elements of the Mathieu 
group :math:`\mbox{M}_{24}`; and we use the numbering of the elements 
of :math:`\mbox{M}_{24}` described in class ``AutPL|`` in module 
``mmgroup``. 

Most functions in module ``mm_group_n.c`` take a pointer to a 5-tuple
:math:`(t, f, e, \delta, \pi)` representing an element :math:`g` 
of the :math:`N_{0}`  as their first argument. Then the tuple 
representing :math:`g` is modified  to a tuple representing an 
element :math:`g_2 = g \cdot g_1`, with the element :math:`g_1` of 
:math:`N` given by one or more subsequent arguments of the function.

The functions in module ``mm_group_n.c`` may cause some overhead due 
to the fact that element of the Mathieu group :math:`\mbox{M}_{24}`
is represented as an integer. But compared to an operation of the 
monster group on its 196884-dimensional that overhead is negligible. 


"""


"""Components of a vector of the representation of the monster 

A vector of the representation of the monster contains 196884
entries which are integers modulo a small odd prime p. The index 
of such an entry is addressed as a tuple (tag, i0, i1).

Here 'tag' is a single capital letter specifying a group of 
indices and i0 and i1 are (usually) integers representing
elements of the Parker loop or the Golay cocode, depending 
on the tag. Details are specified elsewhere.

A tuple  ([scalar], tag, i0, i1) also specifies the multiple of
of scalar * u  of the unit vector u with index (tag, i0, i1).

If such a tuple is given by the user, it is first converted to
of 32-bit integer containing suitable bit fields for its
component scalar, tag, i0, i1. This conversion is done by the 
functions in this module. Indices i0, i1 may also be slices as 
used in numpy. They may also be elements of the Golay code, the 
Parker loop, or the Golay cocode, where appropiate. Furthermore,
an index may be "r" or "n", denoting an index selected at
random.

If a tuple contains one or two slices, it is converted to a
numpy array of type ``uint32``, which contains all entries of the
slice. The one-dimensional array unsigned of 32-bit integers
is passed to the to the fast C routines for reading and updating 
components of a vector.

"""

from numbers import Integral, Number
import warnings
from collections import defaultdict
from collections.abc import Iterable
from functools import partial
from random import randint
from importlib import import_module

import numpy as np

    
try:
    # Try importing the fast C functions
    from mmgroup import mat24 
except (ImportError, ModuleNotFoundError):
    # Use the slow python functions if the C functions are not available
    from mmgroup.dev.mat24.mat24_ref import  Mat24
    mat24 = Mat24


from mmgroup.structures.abstract_rep_space import mod_rand_invertible
from mmgroup.structures.abstract_rep_space import mod_rand_unit

from mmgroup.structures.parse_atoms import ihex 
from mmgroup.structures.gcode import GCode, GcVector 
from mmgroup.structures.cocode import Cocode 
from mmgroup.structures.ploop import PLoop
from mmgroup.structures.suboctad import Octad

ERR_MM_LIN = "Linear access to MM vectors not supported"

try:
    from mmgroup.mm import mm_aux_array_extern_to_sparse
except (ImportError, ModuleNotFoundError):
    warnings.warn(ERR_MM_LIN, UserWarning)   


ERR_MM_SPARSE = "Sparse representation of MM vectors not supported"

try:
    from mmgroup.mm import mm_aux_mul_sparse
except (ImportError, ModuleNotFoundError):
    warnings.warn(ERR_MM_SPARSE, UserWarning)   



########################################################################
# Collecting indices
########################################################################

INDEX_IS_SLICE = 1
INDEX_IS_RANDOM = 2

INDEX_ERROR_TYPE = "%sMM vector index of type '%s' is illegal for tag %s"
INDEX_ERROR_RAND = "Bad random description of MM vector index for tag %s"
INDEX_ERROR_RANGE = "%sMM vector index out if range for tag %s"
                     
POS_NAMES = defaultdict(str)
POS_NAMES.update({1: "First ", 2: "Second "})


########################################################################
# Description of tags
########################################################################




# Ofssets for tags in sparse notation
tag_offsets = {
   "A": 0x2000000,
   "B": 0x4000000,
   "C": 0x6000000,
   "T": 0x8000000,
   "X": 0xa000000,
   "Z": 0xc000000,
   "Y": 0xe000000,
}

# Boundaries for the two indices following a standard tag
tag_lengths = {
        "A": (24,24), "B": (24,24), "C": (24,24), "T":(759,64),
        "X": (2048,24), "Y": (2048,24), "Z": (2048,24), "D": (24,24)
}

# Numpy array dtype used for the sparse representation
U32 = np.uint32

# For each tag, the following dictionary contains the letter for 
# the tag and the formatting functions for the following indices. 
str_tag = {
    0: (None, None, None),
    1: ("A", str, str),
    2: ("B", str, str),
    3: ("C", str, str),
    4: ("T", str, ihex),
    5: ("X", ihex, str),
    6: ("Z", ihex, str),
    7: ("Y", ihex, str),
}


tags = " ABCTXZY"


########################################################################
# Converting numbers and Code elements to integers
########################################################################



def index_iterable(index, max_index, tag, pos):
    """Convert iterable object 'index' to an array of indices

    'index' must be an iterable object containing integers
    0 <= i < 'max_index' only. The function returns a numpy
    array of shape np.uint32 containing these indices in the 
    given order.

    'tag' and 'pos' are used in error messages only. They refer 
    to the tag to which the indes belong, an to the position of
    the indes (0 for the 1st, 1 for the 2nd index).  
    """
    if not all(isinstance(item, Integer) for item in index):
        err = "All MM vector indices in a list must be integers"
        raise TypeError(err)
    if not all(0 <= item < max_index for item in index):
        err = "%sMM vector indices out of range for tag %s"
        raise IndexError(err, POS_NAMES[pos], tag)
    return  INDEX_IS_SLICE, np.array(index, dtype =  U32), 0 


def a_slice(i_slice, max_index):
    """Convert slice object 'i_slice' to an array of indices

    The slice object 'i_slice' is interpreded as in a one-
    dimensional numpy array of length 'max_index'. The function
    returns the indices contained in that slice when applied
    to a one-dimensional numpy array.

    The function returns a numpy arrray of shape np.uint32.
    """
    return np.arange(*i_slice.indices(max_index), dtype = U32)

def index_I(tag, i, pos = 2):
    """Convert an index specifying a bit position

    0 <= 'i' < 24 is an index specifying a bit position in the 
    Golay code. If 'i' is of type Cocode or GcVector, corre-
    sponding to a bit vector of length one, then the position of 
    that bit is taken as the index. The function returns triple

      (index_type, a_indices, sign).

    Here 'index_type' is an integer interpreted as a bit field. 

    If index_type & INDEX_IS_SLICE is not zero then 'i'
    specifies a slice of indices. The numbers of these indices 
    returned in the numpy array 'a_indices'
    
    If index_type & INDEX_IS_SLICE is zero 'i' specifies a
    single index, and the numpy array 'a_indices' of length 1
    contains that indes.

    'sign' is always 0. It is for compatibility with functions 
    converting e.g. a Parker loop element (which comes with
    a natural sign) to a possible negated unit vector.
    
    'pos' ist used in error messages only. It denotes the 
    position of the index (0 for the 1st, 1 for the 2nd index).

    Both indices of tags A, B, and C are of this type and also 
    the second index of tags X, Y and Z. That tag must be given
    as Parameter 'tag'.
    """
    if isinstance(i, Integral):
        if 0  <= i < 24:
            return 0, i, 0
        raise IndexError(INDEX_ERROR_RANGE % (POS_NAMES[pos], tag))
    elif isinstance(i, str):
        if i in ["r", "n"]:
            return INDEX_IS_RANDOM, randint(0,23), 0
        raise IndexError(INDEX_ERROR_RAND % tag)
    elif isinstance(i, slice):     
        return INDEX_IS_SLICE, a_slice(i,24), 0
    elif isinstance(i, Cocode):
        if len(i) == 1: 
            return 0, i.syndrome().bit_list[0], 0  
        err = "MM vector index of class Cocode for tag %d must have weight 1"
        raise IndexError(err % tag)
    elif isinstance(i, GcVector):
        if len(i) == 1: 
            return 0, i.bit_list[0], 0  
        err = "MM vector index of class GcVector for tag %d must have weight 1"
        raise IndexError(err % tag)
    elif isinstance(i, Iterable):
        return index_iterable(i, 24, tag, pos)
    raise TypeError(INDEX_ERROR_TYPE % (POS_NAMES[pos], type(i), tag))


def index_T(tag, i):
    """Convert an index specifying an octad number

    0 <= 'i' < 759 is an index specifying the number of an octad
    in the Golay code. If 'i' is of type GCode or PLoop, corre-
    sponding  to a Golay code word or a Parker Loop element, that 
    element is taken, provided that it (or its complement) has
    length 8; otherwise we raise ValueError. 

    The function returns triple

      (index_type, a_indices, sign).

    Here index_type indicates whether the index has been selected
    ar random, or whether it refers to a slice, as in function 
    index_I().

    The index or the indices making up the slice are returned in
    the numpy array 'a_indices', as in function index_I(). In any
    case the numbers in that array are octad numbers.
    
    Note that a Parker loop element may refer to a negated unit 
    vector. In this case we return the number of the unit vector 
    and we return 'sign' = 1. It is illegal to specify slices 
    of Parker loop elements; so all entries of array 'a_indices'
    have the same sign.

    The first index of tag T is of this type. Parameter 'tag' is
    for compatibility with similar functions; ist must be 'T'.	 
    """
    if isinstance(i, Integral):
        if 0 <= i < 759:
            return 0, i, 0
        raise IndexError(INDEX_ERROR_RANGE % ("First ", tag))
    elif isinstance(i, str):
        if i in ["r", "n"]:
            return INDEX_IS_RANDOM, randint(0,758), 0
        raise IndexError(INDEX_ERROR_RAND % tag)
    elif isinstance(i, slice):     
        return INDEX_IS_SLICE, a_slice(i,759), 0
    elif isinstance(i, (GCode, PLoop)):
        oct = Octad(i) 
        return 0, oct.octad, oct.sign == -1
    elif isinstance(i, Iterable):
        return index_iterable(i, 759, tag, 1)
    raise TypeError(INDEX_ERROR_TYPE % ("First ", type(i), tag))

def index_suboctad(tag, i, index_type_T = 3, a_indices_T = None):
    """Convert an index specifying an suboctad number

    0 <= 'i' < 64 is an index specifying the number of a suboctad
    in the Golay cocode. Here a suboctad is a Golay cocode word
    which can be represented a subset of of an octad. For each
    octad there is a lexicographic numbering of its suboctad, and
    index 'i' refers to that lexicographic numbering. 

    This index type occurs as the second index of tag T only,
    so the first index of that tag gives us a reference octad. 
    If 'i' is of type Cocode or GcVector, corresponding to a 
    Golay cocode word, that cocode word is taken if it fits into 
    the reference octad; otherwise we raise ValueError. 

    The function returns triple

      (index_type, a_indices, sign).

    Here index_type indicates whether the index has been selected
    ar random, or whether it refers to a slice, as in function 
    index_I().

    The index or the indices making up the slice are returned in
    the numpy array 'a_indices', as in function index_I(). In any
    case the numbers in that array are suboctad numbers.

    Parameters ('index_type_T', 'a_indices_T') must be the values 
    (index_type, a_indices) returned by function index_T(), when
    called with the first index for tag T.
    
    The second index of tag T is of this type. Parameter 'tag' is
    for compatibility with similar functions; it must be 'T'.		 
    """
    if isinstance(i, Integral):
        if 0 <= i < 64:
            return 0, i, 0
        raise IndexError(INDEX_ERROR_RANGE % ("Second ", tag))
    elif isinstance(i, str):
        if i in ["r", "n"]:
            return INDEX_IS_RANDOM, randint(0,63), 0
        raise IndexError(INDEX_ERROR_RAND % tag)
    elif isinstance(i, slice):     
        return INDEX_IS_SLICE,  a_slice(i,64), 0
    elif isinstance(i, (GCode, Cocode, GcVector)):
        if index_type_T == 0:
            gcode = mat24.octad_to_gcode(a_indices_T) & 0xfff
            if isinstance(i, GCode):
                cocode = mat24.ploop_cap(gcode, i)
            else:
                cocode = Cocode(i).cocode
            sub = mat24.cocode_to_suboctad(cocode, gcode)
            return 0, sub, 0
        err = "Second vector index for tag T must be integer here"
        raise TypeError(err)
    elif isinstance(i, Iterable):
        return index_iterable(index, 64, tag, 2)
    raise TypeError(INDEX_ERROR_TYPE % ("Second ", type(i), tag))


def index_ploop(tag, i):
    """Convert an index specifying an Parker loop element

    0 <= 'i' < 2**11  is an index specifying the number of an
    element of the Parker loop. The Parker loop has 2**13 
    elements, with the first 2**11 elements being a transversal 
    of the quotient of the Parker Loop by its center. Flipping 
    bit 12 in the number of a Parker loop element always negates 
    the unit vector adressed by that element. Flipping bit 11 may
    either negate or keep that unit vetor, depending on the tag.
    In any case there are 2**11 positive and 2**11 negative unit 
    vectors to be distinguished, and an integer  0 <= 'i' < 2**11
    refers to a positive unit vector.

    The function returns the triple

      (index_type, a_indices, sign).

    Here index_type indicates whether the index has been selected
    ar random, or whether it refers to a slice, as in function 
    index_I().

    The index or the indices making up the slice are returned in
    the numpy array 'a_indices', as in function index_I(). In any
    case the numbers j in that array satisfy 0 <= j < 2**11.

    If 'i' is of type PLoop, corresponding  to a Parker loop 
    element, then the correct (positive or negative) unit vector
    is chosen, and we return 'sign' = 0 for a positve and 
    'sign' = 1 for a negative vector. Note that slices of Parker 
    loop  elements and integers >= 2**11 are illegal, so all 
    entries of the array 'a_indices' have the same sign.

    The first index of tags X, Y and Z is of this type. That tag 
    must be given as Parameter 'tag'.
    """

    if isinstance(i, Integral):
        if 0 <= i < 2048:
            return 0, i, 0
        raise IndexError(INDEX_ERROR_RANGE % ("Second ", tag))
    elif isinstance(i, str):
        if i in ["r", "n"]:
            return INDEX_IS_RANDOM, randint(0,2047), 0
        raise IndexError(INDEX_ERROR_RAND % tag)
    elif isinstance(i, slice):     
        return INDEX_IS_SLICE,  a_slice(i,2048), 0
    elif isinstance(i, (GCode, PLoop)):
        i = i.ord
        sign = (i >> 12) ^ (i >> 11) if tag == "Y" else (i >> 12)
        return 0, i & 0x7ff, sign & 1
    elif isinstance(i, Iterable):
        return index_iterable(index, 2048, tag, 2)
    raise TypeError(INDEX_ERROR_TYPE % ("Second ", type(i), tag))



def index_numeric(tag, i):
    """Convert an numeric index of a vector of the representation

    A vector in the monster group representation may also be 
    adressed by a numeric  index in the same way as a one-
    dimensional numpy array of length 196884.

    The function returns the triple

      (index_type, a_indices, sign).

    Here index_type indicates whether the index has been selected
    ar random, or whether it refers to a slice, as in function 
    index_I().

    The index or the indices making up the slice are returned in
    the numpy array 'a_indices', as in function index_I().

    'sign' is always zero. The 'tag' it for compatibility  with
    similar functions; it should always be 'E'.
    """
    if isinstance(i, Integral):
        if 0 <= i < 196884:
            return 0, i, 0
        raise IndexError(INDEX_ERROR_RANGE % ("First ", tag))
    elif isinstance(i, str):
        if i in ["r", "n"]:
            return INDEX_IS_RANDOM, randint(0,196883), 0
        raise IndexError(INDEX_ERROR_RAND % tag)
    elif isinstance(i, slice):     
        return INDEX_IS_SLICE, a_slice(i, 196884), 0
    elif isinstance(i, Iterable):
        return index_iterable(index, 196884, tag, 1)
    raise TypeError(INDEX_ERROR_TYPE % ("First ", type(i), tag))


########################################################################
# Converting an tuple describing an unit vector to a sparse list 
########################################################################




def rand_str_to_scalar(p, value):
    """Map a character to a random integer modulo p

    For 'value we have:
    'u' means: return 1
    's' means: return 1 or -1 (modulo p)
    'n' means: return a random nonzerodivisor modulo p
    'r' means: return a random number modulo p

    The return value x satisfies 0 <= x < p.
    """
    if value == 'u': return 1
    if value == 's': return mod_rand_unit(p)
    if value == 'n': return mod_rand_invertible(p)
    if value == 'r': return randint(0, p-1)
    err = "Illegal scalar in tuple for MM vector"
    raise TypeError(err)




    


INDEX_ERROR_SLICE = "Slices are illegal in MM vector index tuples"

def gen_unit_A(p, scalar, tag, i0 = 'r', i1 = 'r'):
    """Auxiliary function for function tuple_to_sparse()

    This function deals with tag 'A'. So 'tag' must be 'A'.

    It implements the case for tag "A" where the tag is followd
    by two indices i0, i1. The function returns a numpy array 
    of shape np.uint32 with a single entry:

        (offset << 25) + (i0 << 14) + (i1 << 8) + scalar % p,

    where i0 and i1 are the indices given by the input
    values i0 and i1. 'offset' is given by 'tag_offsets[tag]'. 

    It deals with the following peculiarity:

      -  if i0 < i1:  
             map (tag, i0, i1) to (tag, i1, i0)
    """
    if i0 == 'r' and i1 == None:
        i1 ='r'
    index_type0, i0, _ = index_I(tag, i0, pos = 1)
    index_type1, i1, _ = index_I(tag, i1, pos = 2)
    if (index_type0 | index_type1) & INDEX_IS_SLICE:
        raise TypeError(INDEX_ERROR_SLICE)
    i0, i1 = max(i0, i1), min(i0, i1)
    sc = scalar % p
    a = [tag_offsets[tag] + (i0 << 14) + (i1 << 8) + sc]
    return np.array(a, dtype = U32)
    

def gen_unit_BC(p, scalar, tag, i0 = 'r', i1 = 'r'):
    """Auxiliary function for function tuple_to_sparse()

    This function deals with tags 'B' and 'C'. So 'tag' must
    be 'B' or 'C'.  It implements the case for tags 'B' in
    the same way as function gen_unit_A() implements the
    case for tag 'A'. It deals with the following additional
    peculiarities:

      -  if i0 == i1:
             raise IndexError (This case is illegal)
      -  if i0 or i1 encodes a random index:
             ensure i1 != i0
    """
    if i0 == 'r' and i1 == None:
        i1 ='r'
    index_type0, j0, _ = index_I(tag, i0, pos = 1)
    index_type1, j1, _ = index_I(tag, i1, pos = 2)
    if (index_type0 | index_type1) & INDEX_IS_SLICE:
        raise TypeError(INDEX_ERROR_SLICE)
    is_rnd = (index_type0 | index_type1) & INDEX_IS_RANDOM
    while is_rnd and j0 == j1:
        _, j0, _ = index_I(tag, i0)
        _, j1, _ = index_I(tag, i1)
    if j0 == j1:
        err = "Vector indices 1 and 2 must be different for tag %s"
        raise ValueError(err & tag)
    j0, j1 = max(j0, j1), min(j0, j1)
    sc = scalar % p
    a = [tag_offsets[tag] + (j0 << 14) + (j1 << 8) + sc]
    return np.array(a, dtype = U32)


def gen_unit_D(p, scalar, tag, i0 = 'r', i1 = None):
    """Auxiliary function for function tuple_to_sparse()

    This function deals with tag 'D'. ('D', i) is equivalent
    to ('A', i, i).
    """
    index_type0, i0, _ = index_I(tag, i0, pos = 1)
    if (index_type0) & INDEX_IS_SLICE:
        raise TypeError(INDEX_ERROR_SLICE)
    a = [tag_offsets['A'] + i0 * 0x4100 + scalar % p]
    return np.array(a, dtype = U32)


def gen_unit_T(p, scalar, tag, i0 = 'r', i1 = 'r'):
    """Auxiliary function for function tuple_to_sparse()

    This function deals with tag 'T'. It is implemented in
    the same fashion as function gen_unit_A(). Note that 
    the interpretation of parameter 'i1' depends on 'i0'.
    """
    if i0 == 'r' and i1 == None:
        i1 ='r'
    index_type0, i0, sign = index_T(tag, i0)
    index_type1, i1, _ = index_suboctad(tag, i1, index_type0, i0)
    if (index_type0 | index_type1) & INDEX_IS_SLICE:
        raise TypeError(INDEX_ERROR_SLICE)
    scalar = scalar * (-1)**sign
    a = [tag_offsets[tag] + i0 * 0x4000 + i1 * 0x100 + scalar % p]
    return np.array(a, dtype = U32)


def gen_unit_XYZ(p, scalar, tag, i0 = 'r', i1 = 'r'):
    """Auxiliary function for function tuple_to_sparse()

    This function deals with tags 'X', 'Y' and 'Z'. It is 
    implemented in the same fashion as function gen_unit_A(). 
    Here parameter 'i0' and 'i1' are idependent.
    """
    if i0 == 'r' and i1 == None:
        i1 ='r'
    index_type0, i0, sign = index_ploop(tag, i0)
    index_type1, i1, _ = index_I(tag, i1)
    if (index_type0 | index_type1) & INDEX_IS_SLICE:
        raise TypeError(INDEX_ERROR_SLICE)
    scalar = scalar * (-1)**sign
    a = [tag_offsets[tag] + i0 * 0x4000 + i1 * 0x100 + scalar % p]
    return np.array(a, dtype = U32)


def gen_unit_numeric(p, scalar, tag, i0 = 'r', i1 = None):
    """Auxiliary function for function tuple_to_sparse()

    It deals with tag 'E'. So 'tag' should aways be 'E'. Tag 'E'
    means that 'i0' is interpreted as an index adressing a numpy 
    array of shape (196884,) in the same way as in the C version 
    of the representation of the monster group. 'i0' may be as
    slice in the same way as in a numpy array.
    
    Internally, a vector of the representation of the monster 
    group is not a standard numpy array. We use the C function
    
    mmgroup.structures.mm_aux_array_extern_to_sparse()

    for converting i0 to a standard index.
    """
    index_type0, i0, _ = index_numeric(tag, i0)
    if index_type0 & INDEX_IS_SLICE:
        raise TypeError(INDEX_ERROR_SLICE)
    a = np.array([i0], dtype = U32)
    try:
        mm_aux_array_extern_to_sparse(a, 1)
    except NameError:
        raise NotImplementedError(ERR_MM_LIN)   
    a[0] += scalar % p
    return a
    

def gen_unit_IJ(sign, p, scalar, tag, i0 = 'r', i1 = 'r'):
    """Auxiliary function for function tuple_to_sparse()

    This function deals with tags 'I' and 'J'. Put f(sign, i, j) =

      ('A', i, i) + ('A', j, j) - ('A', i, j) - 2 * sign * ('B', i, j).

    Then ('I', i, j) and ('J', i, j) are equivalent to f(1, i, j) and
    f(-1, i, j), respecvtively, for i != j. The case i == j is illegal.
    """
    def A(i, j, factor = 1):
        sc = scalar * factor % p
        return int(tag_offsets["A"] + (i << 14) + (j << 8) + sc)
    b =  gen_unit_BC(p, -2 * sign * scalar, 'B', i0, i1)[0] 
    i0, i1 = (b >> 14) & 0x1f, (b >> 8) & 0x1f
    a = [A(i0, i0), A(i1, i1), A(i0, i1, -1), b]
    return np.array(a, dtype = U32)



def gen_unit_identity(p, scalar, *args):
    """Auxiliary function for function tuple_to_sparse()

    Generates the identity vector

    scalar * sum([('A', i, i) for i in range(24)])
    """
    start = tag_offsets["A"] + scalar % p
    d = (1 << 14) + (1 << 8)
    return np.array(range(start, start + 24 * d, d), dtype = U32)



def gen_vector_sparse(p, scalar, tag, data, p1=None):
    """Auxiliary function for function sparse_from_indices()

    It deals with tag 'S' where an array 'data' in sparse 
    representation (modulo 'p1') is multiplied with 'scalar'
    converted to a sparse array modulo 'p'.

    Therefore we use function ``mm_aux_mul_sparse``.
    """
    assert tag == 'S'
    if not p1:
        p1 = p
    a = np.array(data, dtype = np.uint32)
    if a.shape == (): 
        a = a.ravel()
    if p1 == p and scalar == 1:
        return a
    a1 = np.zeros(len(a), dtype = np.uint32)
    res =  mm_aux_mul_sparse(p1, a, len(a), scalar, p, a1)
    if res < 0:
        err = "Cannot reduce sparse MM vector module %d"
        raise ValueError(err, p)
    return a1[:res]


def gen_zero(*args):
    return np.zeros(0, dtype = np.uint32)

tuple_to_sparse_functions = {
    'A': gen_unit_A,
    'B': gen_unit_BC,
    'C': gen_unit_BC,
    'T': gen_unit_T,
    'X': gen_unit_XYZ,
    'Z': gen_unit_XYZ,
    'Y': gen_unit_XYZ,
    'D': gen_unit_D,
    'E': gen_unit_numeric,
    'I': partial(gen_unit_IJ, 1), 
    'J': partial(gen_unit_IJ, -1), 
    'U': gen_unit_identity,
    'S': gen_vector_sparse,
    '0': gen_zero,
}



def tuple_to_sparse(p, *data):
    """Convert a tuple to a sparse representation of an MM vector 

    This function maps a tuple to the sparse representation of an
    MM vector, which is (usually) a unit vector. The tuple is 
    preceeded by the first argument p, which is a modulus. 
 
    The general format of such a tuple is 

       (scalar, tag, i0, i1,..)

    Here the 'tag' is a captial letter describing the unit vector,
    i0, i1, ... are indices which are specifying the unit vector,
    depending on the tag. The optional 'scalar' is an integer,
    which is reduced modulo p. It defaults to 1. The 'scalar' may
    also be a lower case letter; in that case a random scalar 
    is computed by function rand_str_to_scalar(p, scalar).

    The function returns the sparse representation of the vector
    to be created as a numpy array with dtype = np.uint32.

    If no data are given, the zero vector is created.
    """ 
    if len(data) == 0:
        return np.array([], dtype = U32)
    scalar = 1
    if isinstance(data[0], Integral):
        scalar, data = data[0] % p, data[1:]
    elif isinstance(data[0], str) and data[0] in "usnr":
        scalar, data = rand_str_to_scalar(p, data[0]), data[1:]
    try:
        tag, indices = data[0], data[1:]
        f = tuple_to_sparse_functions[tag]
    except:
        err = "Illegal or missing tag for creating MM vector"
        raise TypeError(err)
    return f(p, scalar, tag, *indices)


                     
    
        
    



########################################################################
# Converting an tuple describing an index to a sparse list 
########################################################################


SLICE = slice(None)


INDEX_ERROR_SLICE = "Cannot access a random index of an  MM vector"
PLOOP_SIGN_ERROR = "Negative Parker loop elements not allowed as indices" 

def indices_to_sparse(p, tag, type0, i0, type1, i1, sign = 0):
    """Auxiliary function for function sparse_from_indices()

    The function deals with adressing a vector v of the monster
    group representation  in the form v[tag, i0, i1]. Here
    tag, i0, and i1 refer to the index of a (possible negated) 
    component of v. i0 and i1 may also refer to slices of such
    components. 

    A function like index_ploop() or index_I() returns a triple
    (index_type, a_indices, sign) describing the information
    contained in a single index. Here 'index_type' is a bit field
    contaning the following information:

      - Does the index describe a single value or a slice?
      - Has the index been generated at random?  

    The array 'a_indices' is a numpy array containing all
    indices referred by a slice object. It contains one entry
    if the index refers to single component. If 'sign' & 1 == 1
    then the negated components have to be considered instead.

    Let (type0, i0, sign) be the output 
         (index_type, a_indices, sign)
    obtained by applying a function like index_ploop() to the
    first index and let  (type0, i0, dummy) be the output 
         (index_type, a_indices, sign)
    obtained by applying a function like index_I() to the
    second index. Depending on type0 and type1, the
    value v[tag, i0, i1] is either a scalar or an numpy of 
    dimension 1 or 2. 

    The function returns a pair (shape, a_out). Here shape is
    the expected shape of the array v[tag, i0, i1], with 
    shape = () if v[tag, i0, i1] is a scalar. 
    
    a_out is the flattened array contructed from arrays i0 and
    i1 containing the indices (j0, j1) for all j0 in i0 and
    j1 in sparse representation. 

    Remarks:
    
    - The sparse representation corresponding to v[tag, j0, j1] 
      is: 
           tag_offsets[tag]  + (i0 << 14) + (i1 << 8).

    - If 'sign' & 1 == 1 (meaning negated components) then we xor  
      the number p to that sparse representation. This makes sense 
      in case p = 2**k - 1. Then for any x with 0 <= x <= p we have 

           -x  =  x ^ p    (mod p).

      We support negated components in case p == 2**k-1  only.
      
    - If any index i0 or i1 has been generated at random, 
      we raise TypeErrror.

    - We use C-style, i.e. order = C, for reshaping numpy arrays.
    """
    if (type0 | type1) & INDEX_IS_RANDOM:
        raise TypeError(INDEX_ERROR_SLICE)
    offset = tag_offsets[tag] + (p & -sign)
    if (type0 & INDEX_IS_SLICE) == 0:
        if (type1 & INDEX_IS_SLICE) == 0:
            offset += (i1 << 8) + (i0 << 14)            
            return (), np.array([offset], dtype=U32)
        else:
            i1 = np.array(i1,  dtype=U32)
            offset += i0 << 14
            a_out = (i1 << 8) + offset
            return (len(i1),), a_out
    else:
        i0 = np.array(i0,  dtype=U32)
        if (type1 & INDEX_IS_SLICE) == 0:
            offset += i1 << 8
            a_out = (i0 << 14) + offset
            return (len(i0),), a_out
        else:
            i1 = np.array(i1,  dtype=U32)
            shape = (len(i0), len(i1))
            i0 = np.expand_dims(i0, axis=1)
            i1 = np.expand_dims(i1, axis=0)
            a_out = (i0 << 14) + (i1 << 8) + offset
            return shape, a_out.reshape(-1, order = 'C')
    

def ABC_indices_to_sparse(p, tag, i0 = SLICE, i1 = SLICE):
    shape0, a0, _ =  index_I(tag, i0, pos = 1)
    shape1, a1, _ =  index_I(tag, i1, pos = 1)
    return indices_to_sparse(p, tag, shape0, a0, shape1, a1)

def T_indices_to_sparse(p, tag, i0 = SLICE, i1 = SLICE):
    shape0, a0, sign =  index_T(tag, i0)
    shape1, a1, _ =  index_suboctad(tag, i1, shape0, a0)
    if sign and (p & (p+1)):
        raise NotImplementedError(PLOOP_SIGN_ERROR)
    return indices_to_sparse(p, tag, shape0, a0, shape1, a1, sign)

def XYZ_indices_to_sparse(p, tag, i0 = SLICE, i1 = SLICE):
    shape0, a0, sign =  index_ploop(tag, i0)
    shape1, a1, _ =  index_I(tag, i1)
    if sign and (p & (p+1)):
        raise NotImplementedError(PLOOP_SIGN_ERROR)
    return indices_to_sparse(p, tag, shape0, a0, shape1, a1, sign)




def D_index_to_sparse(p, tag, i0 = SLICE):
    shape0, a0, _ =  index_I(tag, i0, pos = 1)
    offset = tag_offsets["A"]
    if shape0 & INDEX_IS_RANDOM:
        raise TypeError(INDEX_ERROR_SLICE)
    if (shape0 and INDEX_IS_SLICE) == 0:
        offset += a0 * 0x4100
        return (), np.array([offset], dtype=U32)
    else:
        a_out = a0 * 0x4100 + offset
        return (len(a0),), a_out
        


def numeric_index_to_sparse(p, tag, i0 = SLICE):
    """Auxiliary function for function sparse_from_indices()

    It deals with tag 'E' which interpretes a single index as
    in the C version of the rep of the monster group.

    Therefore we use function
    
    mmgroup.structures.mm_aux_array_extern_to_sparse()
    """
    shape0, a0, _ =  index_numeric(tag, i0)
    if shape0 & INDEX_IS_RANDOM:
        raise TypeError(INDEX_ERROR_SLICE)
    if (shape0 and INDEX_IS_SLICE) == 0:
        shape = ()
        a_out = np.array([a0], dtype=U32)
    else:
        shape = (len(a0),)
        a_out = a0
    try:
        mm_aux_array_extern_to_sparse(a_out, len(a_out))
    except NameError:
        raise NotImplementedError(ERR_MM_LIN)   
    return shape, a_out




indices_to_sparse_functions = {
    'A': ABC_indices_to_sparse,
    'B': ABC_indices_to_sparse,
    'C': ABC_indices_to_sparse,
    'T': T_indices_to_sparse,
    'X': XYZ_indices_to_sparse,
    'Z': XYZ_indices_to_sparse,
    'Y': XYZ_indices_to_sparse,
    'D': D_index_to_sparse,
    'E': numeric_index_to_sparse,
}


def sparse_from_indices(p, tag, *indices):
    """Convert index of MM vector to sparse representation.

    Although an MM vector is not a numpy ndarray, components 
    of an MM vector can be obtained in the same way as in the
    case of numpy arrays. Here the MM vector v is indexed as 
    v[tag, i0, i1], where 'tag' is a capital letter and 
    i0, i1 are numbers or slices or lists of integers as 
    used in numpy ndarrays.
    
    Here v[tag, i0, i1] returns an integer or a subset of
    the components of v as a numpy array. Coding  
    'v[tag, i0, i1] = A'  means that a number or an array-like 
    object A is converted to a scalar or a numpy array A1 of a 
    shape compatible to the shape of v[tag, i0, i1]. Then the 
    corresponding components of v are updated by those of A1. 

    This function converts the index of an MM vector
    given by (tag, *indices) to a pair

       (shape, a_indices).

    It returns that pair.

    Here 'shape' is a tuple describing the shape of a numpy
    array corresponding to the input (tag, *indices). So if
    the input describes a scalar, we have shape = ().

    'a_indices' is the flat numpy array containing the indices
    described by the input. We use the sparse form (with scalar
    equal to zero) for describing these indices. 

    """
    if isinstance(tag, str):
        return indices_to_sparse_functions[tag](p, tag, *indices)
    return numeric_index_to_sparse(p, 'E', tag, *indices)


########################################################################
# Converting a sparse list entry to a standard form 
########################################################################

def purge_sparse_entry(n):
    """Return the standard form of the sparse index n.

    Return zero if the sparse index n is illegel.

    This is is a slow python function. The fast C functions purge
    sparse indices internally, whenever this is needed.
    """
    tag = n >> 25
    if 1 <= tag <= 3:   # Tags A, B, C
        # Change ('ABC', i0, i1) to ('ABC', i1, i0)  if  i0 < i1
        i0, i1 =  (n >> 14) & 0x7ff, (n >> 8) & 0x3f
        i0, i1 = max(i0, i1),  min(i0, i1) 
        n =  (n & 0xe0000ff) + (i0 << 14) + (i1 << 8)
        # max(i0, i1) >= 24 is illegal for tags 'ABC'
        # i0 == i1 is illegal for tags 'B' and 'C'
        return n if (i0 < 24) and (tag == 1 or i0 != i1) else 0
    elif tag == 4:      # Tag T
        # ('T', i0, i1) is illegal for i0 >= 759
        return n if ((n >> 14) & 0x7ff) < 759 else 0 
    elif 5 <= tag <= 7: # Tags X, Y, Z
        # ('XYZ', i0, i1) is illegal for i1 >= 24
        return n if (n & 0x3f00) < 0x1800 else 0
    else :              # Any other tags are illegal
        return 0

########################################################################
# Converting a sparse list to a list of tuples 
########################################################################

def sparse_to_tuples(a_indices):
    """Convert vector from sparse representation to list of tuples

    Here 'a_indices' is a numpy array representing a vector in
    sparse representation. The function returns a list of tuples

       (scalar, tag, i0, i1)

    Here (tag, i0, i1) represents a unit vector and the integer
    'scalar' is the coordinate of that unit vector.
    """
    return [(i & 0xff, tags[(i >>25) & 7], (i >> 14) & 0x7ff, 
         (i >> 8) & 0x3f) for i in map(int, a_indices) if i & 0xe000000]


########################################################################
# Manipulating indices and data in a sparse list 
########################################################################


def sparse_add_data(p, shape, a_indices, data):
    """Augment indices (given in sparse representation) by data

    Function sparse_from_indices() returns indices in the form
    (shape, a_indices), where 'a_indices' is a flat numpy array
    containing the indices selected from an MM vector v. Here
    'shape' is the shape of that array.

    When updating these indices of v with data, we first augment 
    the array 'a_indices' with the arrray-like object 'data', 
    containing the data used for updating. Here the entries of 
    'data' are reduced modulo the characteristic 'p' of the 
    vector space.
    
    The array-like object 'data' is converted to a numpy array
    with the given 'shape'.
    """
    if isinstance(data, Number):
        a_indices ^= data % p
    else:
        if not isinstance(data, np.ndarray):
            data = np.array(data, dtype = np.int64)
        data = (data % p).astype(U32)
        if data.shape != shape:
            err = "Shape mismatch when updating mm vector"
            raise ValueError(err)
        a_indices ^= data.reshape(-1)
    return a_indices
    





     
def sparse_to_ndarray(p, shape, a_indices):
    """Convert vector in sparse representation to array

    This function returns the data contained in the vector
    'a_indices' in sparse representation to an numpy ndarray of 
    given 'shape'. The entries of the vector are reduced 
    modulo  p.   
    """
    a = np.reshape((a_indices & 0xff) % p, shape).astype(np.uint8)
    return a if len(shape) else int(a)


def sparse_to_str(p, a_indices):
    """Convert vector in sparse representation to a string

    This function returns the data contained in the vector
    'a_indices' in sparse representation as a string. The 
    entries of the vector are reduced modulo p.   
    """
    names = []
    half = p >> 1
    for a in a_indices:
        tag, fmt0, fmt1 = str_tag[(a >> 25) & 7]
        i0, i1, v = (a >> 14) & 0x7ff, (a >> 8) & 0x3f, (a & 0xff) % p
        if tag and v:
            s, vabs = ("-", p - v) if v > half else ("+", v)
            s += (str(vabs) + '*') if vabs > 1 else ""
            names.append("%s%s_%s_%s" % (s, tag, fmt0(i0), fmt1(i1)))            
    if len(names): 
        if names[0][:1] == "+": 
            names[0] = names[0][1:]  
        return "".join(names)
    return "0"

    